﻿Public Class EditorFormNew

    Dim Video As Roton.VideoGDI
    Dim Environment As Roton.Environment
    Dim FontData() As Byte
    Dim Input As Roton.InputForm

    Public Sub New(ByVal NewEnvironment As Roton.Environment, ByVal NewFont() As Byte)
        ' This call is required by the designer.
        InitializeComponent()
        Application.DoEvents()

        ' Add any initialization after the InitializeComponent() call.
        Environment = NewEnvironment
        FontData = NewFont

        ' set form style
        SetStyle(ControlStyles.AllPaintingInWmPaint, True)
        SetStyle(ControlStyles.ResizeRedraw, True)
        SetStyle(ControlStyles.UserPaint, True)
        Me.Text = "Editor - " & Environment.World.Name

        ' set up video
        Video = New Roton.VideoGDI
        Video.SetFont(FontData)
        Video.SetViewport(80, 25)
        Environment.Video = Video
        RefreshMenu()

        ' set environment for editor mode
        Environment.SetEditorMode(True)

        ' set up input
        Input = New Roton.InputForm
        Environment.Input = Input

        ' set up scroll
        Environment.Scroll = New Roton.ScrollVideo(Video, Environment.State, Environment.Parameters)

        ' set board
        Environment.SetBoard(Environment.World.Board)
        Environment.Execute()
        Video.Update()
        BackgroundImage = Video.Bitmap
        ClientSize = BackgroundImage.Size
    End Sub

    Public Sub RefreshMenu()
        Dim X As Integer
        Dim Y As Integer
        For Y = 0 To 24
            For X = 60 To 79
                Video.PlotOverlay(X, Y, New Roton.AnsiChar(32, &H1F))
            Next
        Next

        ' location
        Video.WriteOverlay(63, 0, "(30, 13) 0/150", &H1B)

        ' functions
        Video.WriteOverlay(61, 3, " L ", &H30)
        Video.WriteOverlay(65, 3, "Load", &H1E)
        Video.WriteOverlay(70, 3, " H ", &H30)
        Video.WriteOverlay(74, 3, "Help", &H1E)
        Video.WriteOverlay(61, 4, " S ", &H70)
        Video.WriteOverlay(65, 4, "Save", &H1E)
        Video.WriteOverlay(70, 4, " Q ", &H70)
        Video.WriteOverlay(74, 4, "Quit", &H1E)
        Video.WriteOverlay(61, 6, " I ", &H30)
        Video.WriteOverlay(65, 6, "Info", &H1E)
        Video.WriteOverlay(70, 6, " W ", &H30)
        Video.WriteOverlay(74, 6, "World", &H1E)
        Video.WriteOverlay(61, 7, " B ", &H70)
        Video.WriteOverlay(65, 7, "Switch Board", &H1E)

        ' categories
        Video.WriteOverlay(61, 9, "  F1   ", &H30)
        Video.WriteOverlay(69, 9, "Item", &H1E)
        Video.WriteOverlay(61, 10, "  F2   ", &H70)
        Video.WriteOverlay(69, 10, "Creature", &H1E)
        Video.WriteOverlay(61, 11, "  F3   ", &H30)
        Video.WriteOverlay(69, 11, "Terrain", &H1E)
        Video.WriteOverlay(61, 12, "  F4   ", &H70)
        Video.WriteOverlay(69, 12, "Enter Text", &H1E)

        ' controls
        Video.WriteOverlay(61, 14, " Space ", &H30)
        Video.WriteOverlay(69, 14, "Plot", &H1E)
        Video.WriteOverlay(61, 15, "  Tab  ", &H70)
        Video.WriteOverlay(69, 15, "Draw", &H1E)

        ' pattern
        Video.WriteOverlay(61, 17, " P ", &H30)
        Video.WriteOverlay(65, 17, "Pattern", &H1E)
        Video.WriteOverlay(61, 18, " C ", &H70)
        Video.WriteOverlay(65, 18, "Color", &H1E)
        Video.WriteOverlay(61, 19, " V ", &H30)
        Video.WriteOverlay(65, 19, "Blink", &H1E)

        ' color selector
        For X = 0 To 7
            Video.PlotOverlay(61 + X, 23, New Roton.AnsiChar(223, X))
            Video.PlotOverlay(69 + X, 23, New Roton.AnsiChar(223, (X + 8) Or ((X And 7) << 4)))
        Next

        ' color selector arrows
        Video.PlotOverlay(61, 22, New Roton.AnsiChar(31, &H1F))
        Video.PlotOverlay(69, 24, New Roton.AnsiChar(30, &H1F))
    End Sub

    Protected Overrides Sub OnPaintBackground(ByVal e As System.Windows.Forms.PaintEventArgs)
        Dim DisplayWidth As Integer
        Dim DisplayHeight As Integer
        Dim DisplayRatio As Double = ClientSize.Width / ClientSize.Height
        Dim ImageRatio As Double = (BackgroundImage.Width) / (BackgroundImage.Height)
        Dim DisplayX As Integer
        Dim DisplayY As Integer

        If (DisplayRatio > ImageRatio) Then
            DisplayHeight = ClientSize.Height
            DisplayWidth = DisplayHeight * ImageRatio
        Else
            DisplayWidth = ClientSize.Width
            DisplayHeight = DisplayWidth / ImageRatio
        End If

        DisplayX = (ClientSize.Width - DisplayWidth) \ 2
        DisplayY = (ClientSize.Height - DisplayHeight) \ 2

        With e.Graphics
            .Clear(Color.FromArgb(255, 24, 24, 24))
            If (mnuViewInterpolated.Checked) AndAlso ((ClientSize.Width <> DisplayWidth) OrElse (ClientSize.Height <> DisplayHeight)) Then
                .InterpolationMode = Drawing2D.InterpolationMode.Low
            Else
                .InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
            End If
            .PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
            .DrawImage(BackgroundImage, DisplayX, DisplayY, DisplayWidth, DisplayHeight)
        End With
    End Sub

End Class