﻿Public Class EditorForm
    Private Shared ReadOnly GameEncoding As System.Text.Encoding = System.Text.Encoding.GetEncoding(437)
    Private Shared ReadOnly TextEncoding As System.Text.Encoding = System.Text.Encoding.Unicode

    Public Environment As Roton.Environment
    Private Video As Roton.VideoGDI
    Private Input As Roton.InputForm
    Private LastBoard As Integer = -1
    Private Board As Roton.Board
    Private Thing As Roton.Thing
    Private FontData() As Byte
    Private ActiveTileButton As TSButton
    Private ActiveTileImage As Bitmap
    Private BufferButton As TSButton
    Private StatsButton As TSButton
    Private DefaultColorButton As TSButton
    Private HexButton As TSButton
    Private BufferImage As Bitmap
    Private Buffer() As BufferItem
    Private TSHandler As New MouseEventHandler(AddressOf ButtonDown)
    Private TSColorHandler As New MouseEventHandler(AddressOf ColorDown)
    Private TileWidth As Integer
    Private TileHeight As Integer
    Private ViewTileWidth As Integer
    Private ViewTileHeight As Integer
    Private MouseX As Integer
    Private MouseY As Integer
    Private LastMouseX As Integer
    Private LastMouseY As Integer
    Private Status As ToolStripStatusLabel
    Private Initialized As Boolean = False
    Private P3IsBoard As Boolean
    Private SelectedColor As Integer = &HF
    Private CharImages(255) As Image
    Private BoardMenuIndex As Integer

    Private Structure Coordinate
        Public X As Integer
        Public Y As Integer
        Public Sub New(ByVal NewX As Integer, ByVal NewY As Integer)
            X = NewX
            Y = NewY
        End Sub
    End Structure

    Private Structure BufferItem
        Public Stats As Boolean
        Public Tile As Roton.Tile
        Public Thing As Roton.Thing
    End Structure

    Private Class TSButton
        Inherits ToolStripButton

        Public Sub New(ByVal NewName As String, ByVal NewText As String, ByVal MouseDownHandler As MouseEventHandler)
            MyBase.New(NewText)
            DisplayStyle = ToolStripItemDisplayStyle.Text
            Name = NewName
            AddHandler MouseDown, MouseDownHandler
        End Sub

        Public Sub New(ByVal NewName As String, ByVal NewText As String, ByVal NewColor As Color, ByVal MouseDownHandler As MouseEventHandler)
            MyBase.New(NewText)
            BackColor = NewColor
            With NewColor
                If ((CInt(.R) + CInt(.G) + CInt(.B)) \ 3 >= &H80) Then
                    ForeColor = Color.Black
                Else
                    ForeColor = Color.White
                End If
            End With
            DisplayStyle = ToolStripItemDisplayStyle.Text
            Name = NewName
            AddHandler MouseDown, MouseDownHandler
        End Sub

        Public Sub New(ByVal NewName As String, ByVal NewImage As Image, ByVal MouseDownHandler As MouseEventHandler)
            MyBase.New(NewImage)
            ImageScaling = ToolStripItemImageScaling.None
            DisplayStyle = ToolStripItemDisplayStyle.Image
            Name = NewName
            AddHandler MouseDown, MouseDownHandler
        End Sub

        'Protected Overrides Sub OnMouseDown(ByVal e As System.Windows.Forms.MouseEventArgs)
        '    MyBase.OnMouseDown(e)
        'End Sub

        'Protected Overrides Sub OnMouseHover(ByVal e As System.EventArgs)
        '    'MyBase.OnMouseHover(e)
        'End Sub

        'Protected Overrides Sub OnMouseEnter(ByVal e As System.EventArgs)
        '    'MyBase.OnMouseEnter(e)
        'End Sub

        'Protected Overrides Sub OnMouseLeave(ByVal e As System.EventArgs)
        '    'MyBase.OnMouseLeave(e)
        'End Sub

        'Protected Overrides Sub OnMouseMove(ByVal e As System.Windows.Forms.MouseEventArgs)
        '    'MyBase.OnMouseMove(e)
        'End Sub

        'Protected Overrides Sub OnClick(ByVal e As System.EventArgs)
        '    'MyBase.OnClick(e)
        'End Sub

        'Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
        '    MyBase.OnPaint(e)
        'End Sub
    End Class

    Public Sub New(ByVal NewEnvironment As Roton.Environment, ByVal NewFont() As Byte, ByVal StatusLabel As ToolStripStatusLabel)
        ' This call is required by the designer.
        InitializeComponent()
        Application.DoEvents()

        ' Add any initialization after the InitializeComponent() call.
        Status = StatusLabel
        Environment = NewEnvironment
        FontData = NewFont

        ' set up video
        Video = New Roton.VideoGDI
        Video.SetFont(FontData)
        Video.SetViewport(Environment.Parameters.BoardWidth, Environment.Parameters.BoardHeight)
        Environment.Video = Video
        InitViewport()

        ' set up chars
        InitChars()

        ' set up colors
        InitColors()

        ' set up buffer
        InitBuffer()
        UpdateBuffer()

        ' fill info
        CreateTrees()
        LoadWorldInfo()

        ' set environment for editor mode
        Environment.SetEditorMode(True)
        SetBoard(Environment.World.Board)

        ' set up views
        Select Case Environment.Type
            Case Roton.EnvironmentGameType.ZZT
                txtWorldFlagA.Enabled = False
                txtWorldFlagB.Enabled = False
                txtWorldFlagC.Enabled = False
                txtWorldFlagD.Enabled = False
                txtWorldFlagE.Enabled = False
                txtWorldFlagF.Enabled = False
                numWorldStones.Enabled = False
            Case Roton.EnvironmentGameType.SuperZZT
                chkBoardDark.Enabled = False
                numWorldTorches.Enabled = False
                numWorldTorchCycles.Enabled = False
        End Select
        numThingX.Enabled = False
        numThingY.Enabled = False

        'picMain.Image = Video.Bitmap
        Initialized = True
        tmrMain.Enabled = True
        picMain.Visible = True
    End Sub

    Private Sub InitViewport()
        Dim FinalSize As Size
        With Video.Bitmap
            If (Environment.Parameters.DisplayWide) Then
                FinalSize = New Size(.Width * 2, .Height)
            Else
                FinalSize = .Size
            End If
        End With
        TileWidth = FinalSize.Width \ Video.ViewWidth
        TileHeight = FinalSize.Height \ Video.ViewHeight
        If (mnuView2x.Checked) Then
            FinalSize.Width *= 2
            FinalSize.Height *= 2
        End If
        ViewTileWidth = FinalSize.Width \ Video.ViewWidth
        ViewTileHeight = FinalSize.Height \ Video.ViewHeight
        picMain.ClientSize = FinalSize
    End Sub

    Private Sub InitChars()
        Dim Index As Integer
        Dim TSI As ToolStripMenuItem

        mnuCharList.SuspendLayout()
        mnuHexColor.SuspendLayout()
        For Index = 0 To 255
            CharImages(Index) = Video.ExportTile(New Roton.AnsiChar(Index, &HF))
            TSI = New ToolStripMenuItem(CharImages(Index))
            TSI.ImageScaling = ToolStripItemImageScaling.None
            TSI.Text = Index & " (" & Hex(Index) & "h)"
            mnuCharList.Items.Add(TSI)
            TSI = New ToolStripMenuItem(CharImages(Index))
            TSI.ImageScaling = ToolStripItemImageScaling.None
            TSI.Text = Index & " (" & Hex(Index) & "h)"
            mnuHexColor.Items.Add(TSI)
        Next
        mnuCharList.ResumeLayout()
        mnuHexColor.ResumeLayout()
    End Sub

    Private Sub InitColors()
        Dim Index As Integer
        Dim Pal As Color() = Video.Palette

        tsColors.SuspendLayout()
        HexButton = New TSButton("Hex", "", TSHandler)
        tsColors.Items.Add(HexButton)
        For Index = 0 To 15
            tsColors.Items.Add(New TSButton("Color" & Index, Hex(Index), Pal(Index), TSColorHandler))
        Next
        tsColors.ResumeLayout()
    End Sub

    Private Sub InitBuffer()
        ReDim Buffer(8)
        Dim Index As Integer

        tsColors.SuspendLayout()
        For Index = 0 To 8
            Buffer(Index) = New BufferItem
            With Buffer(Index)
                .Tile = New Roton.Tile(Environment.Library.GetElementFromIndex(0), &HF)
                .Thing = New Roton.Thing
                .Thing.UnderElement = Environment.Library.GetElementFromID(Roton.ElementID.Empty)
                .Stats = False
            End With
        Next

        ActiveTileButton = New TSButton("Active", ActiveTileImage, TSHandler)
        ActiveTileButton.ImageScaling = ToolStripItemImageScaling.None
        tsColors.Items.Add(ActiveTileButton)

        StatsButton = New TSButton("Stats", "Stats", TSHandler)
        tsColors.Items.Add(StatsButton)

        DefaultColorButton = New TSButton("Default", "Default", TSHandler)
        DefaultColorButton.Checked = False
        tsColors.Items.Add(DefaultColorButton)

        BufferButton = New TSButton("Buffer", BufferImage, TSHandler)
        BufferButton.ImageScaling = ToolStripItemImageScaling.None
        tsColors.Items.Add(BufferButton)

        tsColors.ResumeLayout()
    End Sub

    Private Sub UpdateBuffer()
        Dim Index As Integer
        Dim G As Graphics

        tsColors.SuspendLayout()

        BufferImage = New Bitmap(TileWidth * 8, TileHeight, Imaging.PixelFormat.Format32bppPArgb)
        ActiveTileImage = New Bitmap(TileWidth, TileHeight, Imaging.PixelFormat.Format32bppPArgb)

        If (DefaultColorButton.Checked) Then
            Buffer(0).Tile.Color = Environment.EditorChar(SelectedColor, Buffer(0).Tile.Element).Color
        Else
            Buffer(0).Tile.Color = SelectedColor
        End If

        G = Graphics.FromImage(BufferImage)
        G.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
        G.PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
        For Index = 0 To 7
            G.DrawImage(Video.ExportTile(GetChar(Buffer(Index + 1))), Index * TileWidth, 0, TileWidth, TileHeight)
        Next

        G = Graphics.FromImage(ActiveTileImage)
        G.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
        G.PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
        G.DrawImage(Video.ExportTile(GetChar(Buffer(0))), 0, 0, TileWidth, TileHeight)

        ActiveTileButton.Image = ActiveTileImage
        BufferButton.Image = BufferImage
        StatsButton.Checked = Buffer(0).Stats
        HexButton.Text = Hex(SelectedColor)
        If (HexButton.Text.Length = 1) Then
            HexButton.Text = "0" & HexButton.Text
        End If

        tsColors.ResumeLayout()
    End Sub

    Private Function GetChar(ByVal B As BufferItem) As Roton.AnsiChar
        Return Environment.RenderedChar(B.Tile, B.Thing, 15)
    End Function

    Private Sub UpdateBoardLists()
        Dim BoardNames As New List(Of String)
        Dim BoardName As String
        Dim Index As Integer

        BoardTree.SuspendLayout()
        cmbBoardExitN.SuspendLayout()
        cmbBoardExitS.SuspendLayout()
        cmbBoardExitW.SuspendLayout()
        cmbBoardExitE.SuspendLayout()
        cmbWorldStart.SuspendLayout()

        BoardTree.Nodes.Clear()
        For Index = 0 To Environment.Boards.Count - 1
            BoardName = Environment.Boards(Index).Name
            BoardTree.Nodes.Add(BoardName)
            BoardNames.Add(BoardName)
        Next

        Index = cmbWorldStart.SelectedIndex
        cmbWorldStart.Items.Clear()
        cmbBoardExitN.Items.Clear()
        cmbBoardExitS.Items.Clear()
        cmbBoardExitW.Items.Clear()
        cmbBoardExitE.Items.Clear()

        cmbWorldStart.Items.AddRange(BoardNames.ToArray)
        BoardNames(0) = "-"

        cmbBoardExitN.Items.AddRange(BoardNames.ToArray)
        cmbBoardExitS.Items.AddRange(BoardNames.ToArray)
        cmbBoardExitW.Items.AddRange(BoardNames.ToArray)
        cmbBoardExitE.Items.AddRange(BoardNames.ToArray)
        cmbBoardExitN.SelectedIndex = 0
        cmbBoardExitS.SelectedIndex = 0
        cmbBoardExitW.SelectedIndex = 0
        cmbBoardExitE.SelectedIndex = 0
        If (Index < 0) Then
            cmbWorldStart.SelectedIndex = Environment.World.Board
        Else
            cmbWorldStart.SelectedIndex = Index
        End If
        BoardTree.SelectedNode = BoardTree.Nodes(Environment.World.Board)

        BoardTree.ResumeLayout()
        cmbBoardExitN.ResumeLayout()
        cmbBoardExitS.ResumeLayout()
        cmbBoardExitW.ResumeLayout()
        cmbBoardExitE.ResumeLayout()
        cmbWorldStart.ResumeLayout()

    End Sub

    Private Sub UpdateThingList()
        Dim Index As Integer
        Dim Entry As String
        Dim N As String
        Dim Items As New List(Of String)

        ThingsTree.SuspendLayout()
        mnuThingBind.SuspendLayout()
        mnuRC.SuspendLayout()
        mnuRCBind.Visible = False

        ThingsTree.Nodes.Clear()
        mnuThingBind.Items.Clear()
        mnuRCBind.DropDownItems.Clear()

        For Index = 0 To Board.Things.Count - 1
            With Board.Things(Index)
                Entry = "[" & .X & "," & .Y & "] " & Board.ElementAt(.X, .Y).ID.ToString
                If (Not (.Code Is Nothing)) Then
                    If (.Code.Code.StartsWith("@")) AndAlso (.Code.Code.Length > 1) Then
                        N = .Code.Code.Substring(1)
                        If (N.Contains(Chr(13))) Then
                            N = N.Substring(0, N.IndexOf(Chr(13)))
                        End If
                    Else
                        N = ""
                    End If
                    If (.Length < 0) Then
                        Entry &= "*"
                    End If
                    Entry &= " """ & N & """"
                End If
                ThingsTree.Nodes.Add(New TreeNode(Entry))
                mnuThingBind.Items.Add(Entry)
                mnuRCBind.DropDownItems.Add(Entry)
                If (.Length < 0) OrElse (.Code Is Nothing) OrElse (Not .Code.Code.StartsWith("@")) Then
                    mnuThingBind.Items(mnuThingBind.Items.Count - 1).Visible = False
                    mnuRCBind.DropDownItems(mnuRCBind.DropDownItems.Count - 1).Visible = False
                End If
            End With
        Next

        mnuThingBind.Items.RemoveAt(0)
        mnuThingBind.Items.Insert(0, New ToolStripMenuItem("(none)"))

        mnuRCBind.Visible = True
        ThingsTree.ResumeLayout()
        mnuThingBind.ResumeLayout()
        mnuRC.ResumeLayout()
    End Sub

    Private Sub CreateTrees()
        Dim Index As Integer
        Dim L As Roton.Library = Environment.Library
        Dim E As Roton.Element
        Dim Count As Integer = L.Count
        Dim ThisNode As TreeNode
        Dim Tip As String
        Dim TargetMenu As String

        ElementTree.SuspendLayout()

        ElementTree.Nodes.Clear()
        ElementTree.Nodes.Add("1", "Items")
        ElementTree.Nodes.Add("2", "Creatures")
        ElementTree.Nodes.Add("3", "Terrain")
        ElementTree.Nodes.Add("4", "Nasties")
        ElementTree.Nodes.Add("5", "Misc")
        ElementTree.Nodes.Add("6", "Text")
        ElementTree.Nodes.Add("0", "Unlisted")
        ElementTree.ShowNodeToolTips = True
        cmbThingUnderID.Items.Clear()

        For Index = 0 To Count - 1
            E = L.GetElementFromIndex(Index)
            cmbThingUnderID.Items.Add(E.ID.ToString)
            With E
                TargetMenu = .Menu.ToString
                If (TargetMenu = "5") Then
                    ' group terrain and misc into one
                    TargetMenu = "3"
                End If
                ThisNode = ElementTree.Nodes(TargetMenu).Nodes.Add("E" & Index.ToString, .ID.ToString)
                Tip = "Element " & Hex(Index) & "h" & vbCrLf
                Select Case E.ID
                    Case Roton.ElementID.Ammo
                        Tip &= "Ammunition for the player to shoot."
                    Case Roton.ElementID.Bear
                        Tip &= "Bears charge the player based on alignment and sensitivity."
                    Case Roton.ElementID.BlinkWall
                        Tip &= "Blink walls shoot harmful rays in one direction."
                    Case Roton.ElementID.BoardEdge
                        Tip &= "Board edge acts the same way the real edge of the board does."
                    Case Roton.ElementID.Bomb
                        Tip &= "Bombs can be activated by the player, and count down to an explosion."
                    Case Roton.ElementID.Boulder
                        Tip &= "Boulders are obstacles that can be pushed."
                    Case Roton.ElementID.Breakable
                        Tip &= "Breakable walls can be destroyed by bears, bullets and stars."
                    Case Roton.ElementID.Bullet
                        Tip &= "Bullets travel in one direction and are harmful."
                    Case Roton.ElementID.Clockwise, Roton.ElementID.Counter
                        Tip &= "Conveyors manipulate tiles around them in a circular motion."
                    Case Roton.ElementID.Door
                        Tip &= "Doors require keys to pass through."
                    Case Roton.ElementID.DragonPup
                        Tip &= "Dragon pups are stationary and are only harmful if touched."
                    Case Roton.ElementID.Duplicator
                        Tip &= "Duplicators create copies of items and creatures behind them."
                    Case Roton.ElementID.Empty
                        Tip &= "An empty space."
                    Case Roton.ElementID.Energizer
                        Tip &= "Energizers give the player immunity to creatures and projectiles."
                    Case Roton.ElementID.Fake
                        Tip &= "Fake walls can be walked through, but nothing can be pushed over them."
                    Case Roton.ElementID.Floor
                        Tip &= "Floors can be walked on, but nothing can be pushed on them."
                    Case Roton.ElementID.Forest
                        Tip &= "Forests block projectiles and creatures, but can be cleared by walking through."
                    Case Roton.ElementID.Gem
                        Tip &= "Gems are often used as currency in games. Additionally, they provide health."
                    Case Roton.ElementID.Head
                        Tip &= "Heads are the brains of centipede. Combine with segments for longer centipedes."
                    Case Roton.ElementID.Invisible
                        Tip &= "Invisible walls become normal walls when the player runs into them."
                    Case Roton.ElementID.Key
                        Tip &= "Keys are required to open doors."
                    Case Roton.ElementID.Lava
                        Tip &= "Lava is impassible by the player, but bullets and stars can pass through freely."
                    Case Roton.ElementID.Line
                        Tip &= "Lines are walls that change appearance based on other nearby lines."
                    Case Roton.ElementID.Lion
                        Tip &= "Lions are the most basic creature. They wander toward the player."
                    Case Roton.ElementID.Normal
                        Tip &= "Normal walls are just regular walls."
                    Case Roton.ElementID.Object
                        Tip &= "Objects may contain scripts that perform many interesting tasks on the board."
                    Case Roton.ElementID.Pairer
                        Tip &= "Pairers are stationary and do nothing, but are harmful if walked into."
                    Case Roton.ElementID.Passage
                        Tip &= "Passages can lead to other boards."
                    Case Roton.ElementID.Player
                        Tip &= "Players are directly controlled by user input. Placing this element will create clones."
                    Case Roton.ElementID.Pusher
                        Tip &= "Pushers push things in one direction constantly."
                    Case Roton.ElementID.Ricochet
                        Tip &= "Ricochets make bullets change direction."
                    Case Roton.ElementID.Roton
                        Tip &= "Rotons run circles around the player to make it difficult to fight back."
                    Case Roton.ElementID.Ruffian
                        Tip &= "Ruffians move in short, quick bursts and rest in between."
                    Case Roton.ElementID.Scroll
                        Tip &= "Scrolls display a message on the screen when touched."
                    Case Roton.ElementID.Segment
                        Tip &= "Segments are extensions to centipedes. They will become Heads if separated."
                    Case Roton.ElementID.Shark
                        Tip &= "Sharks move around only in the water. They will attack nearby players."
                    Case Roton.ElementID.SliderEW
                        Tip &= "Pushable slider that only moves east-west."
                    Case Roton.ElementID.SliderNS
                        Tip &= "Pushable slider that only moves north-south."
                    Case Roton.ElementID.Slime
                        Tip &= "Slimes multiply, leaving breakable walls in their path."
                    Case Roton.ElementID.Solid
                        Tip &= "Solid walls are just regular walls."
                    Case Roton.ElementID.Spider
                        Tip &= "Spiders only travel on webs and have very jittery movement."
                    Case Roton.ElementID.SpinningGun
                        Tip &= "Spinning guns shoot bullets randomly in all four directions."
                    Case Roton.ElementID.Star
                        Tip &= "Stars track the player and can only be destroyed with a bomb or when the star expires."
                    Case Roton.ElementID.Stone
                        Tip &= "Stones of power can be used as a currency in Super ZZT games."
                    Case Roton.ElementID.Tiger
                        Tip &= "Tigers try to move toward the player, and can shoot bullets or stars."
                    Case Roton.ElementID.Torch
                        Tip &= "Torches are used to light up dark boards."
                    Case Roton.ElementID.Transporter
                        Tip &= "Transporters can move things to other parts of the board. Use facing transporters for long range movement."
                    Case Roton.ElementID.Water
                        Tip &= "Water is impassible to the player but bullets and stars may cross."
                    Case Roton.ElementID.WaterE, Roton.ElementID.WaterN, Roton.ElementID.WaterS, Roton.ElementID.WaterW
                        Tip &= "Rivers can be made to move the player based on the direction of its arrow."
                    Case Roton.ElementID.Web
                        Tip &= "Webs can be moved through with ease, but they will stick to whatever moves through."
                End Select
                ThisNode.ToolTipText = Tip
            End With
        Next
        UpdateBoardLists()

        ' remove all categories that are empty
        Do
            For Index = 0 To ElementTree.Nodes.Count - 1
                If (ElementTree.Nodes(Index).Nodes.Count = 0) Then
                    ElementTree.Nodes(Index).Remove()
                    Continue Do
                End If
            Next
            Exit Do
        Loop
        ElementTree.Sort()

        ElementTree.ResumeLayout()
    End Sub

    Private Sub SetTitle()
        Text = Environment.World.Name & " - [" & Environment.World.Board & "] " & Environment.CurrentBoard.Name
    End Sub

    Private Sub tmrMain_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles tmrMain.Tick
        If (Not (Environment Is Nothing)) AndAlso (Environment.Execute) Then
            Video.Update()
            picMain.Invalidate()
        End If
        If (Environment Is Nothing) OrElse (Environment.State.QuitZZT) Then
            Close()
        ElseIf (LastBoard <> Environment.World.Board) Then
            SetTitle()
        End If
    End Sub

    Private Sub picMain_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picMain.MouseDown
        If (Not Initialized) Then
            Exit Sub
        End If
        UpdateMouse(e)
        Select Case e.Button
            Case Windows.Forms.MouseButtons.Left
                PlotTile(MouseX, MouseY)
            Case Windows.Forms.MouseButtons.Middle
                GrabTile(MouseX, MouseY)
            Case Windows.Forms.MouseButtons.Right
                mnuRCBind.Enabled = (Board.ThingIndexAt(MouseX, MouseY) >= 0)
                mnuRC.Show(picMain, e.X, e.Y)
        End Select
    End Sub

    Private Sub picMain_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picMain.MouseMove
        If (Not Initialized) Then
            Exit Sub
        End If
        UpdateMouse(e)
        Select Case e.Button
            Case Windows.Forms.MouseButtons.Left
                If (MouseX <> LastMouseX) OrElse (MouseY <> LastMouseY) Then
                    PlotTile(MouseX, MouseY)
                End If
        End Select
    End Sub

    Private Sub UpdateMouse(ByVal e As System.Windows.Forms.MouseEventArgs)
        Dim StatusText As String
        Dim Index As Integer

        LastMouseX = MouseX
        LastMouseY = MouseY
        MouseX = (e.X \ ViewTileWidth)
        MouseY = (e.Y \ ViewTileHeight)
        StatusText = "[" & MouseX & ", " & MouseY & "] " & Board.ElementAt(MouseX, MouseY).ID.ToString
        Index = Board.ThingIndexAt(MouseX, MouseY)
        If (Index >= 0) Then
            StatusText &= " #" & Index
        End If
        Status.Text = StatusText
    End Sub

    Private Sub picMain_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picMain.MouseUp
        If (Not Initialized) Then
            Exit Sub
        End If
        Select Case e.Button
            Case Windows.Forms.MouseButtons.Left
                UpdateThingList()
        End Select
    End Sub

    Private Sub picMain_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles picMain.Paint
        With e.Graphics
            .InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
            .PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
            .DrawImage(Video.Bitmap, picMain.ClientRectangle)
            .DrawRectangle(New Pen(Brushes.White), New Rectangle(MouseX * ViewTileWidth, MouseY * ViewTileHeight, ViewTileWidth + 1, ViewTileHeight + 1))
            If (tabMain.SelectedTab.Equals(tabThings)) AndAlso (Not (Thing Is Nothing)) Then
                .DrawRectangle(New Pen(Brushes.Orange), New Rectangle(Thing.X * ViewTileWidth, Thing.Y * ViewTileHeight, ViewTileWidth + 1, ViewTileHeight + 1))
            End If
        End With
    End Sub

    'Private Sub BoardTree_AfterSelect(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles BoardTree.AfterSelect
    '    SetBoard(e.Node.Index)
    'End Sub

    Private Sub AddBuffer(ByVal NewThing As Roton.Thing, ByVal NewTile As Roton.Tile, ByVal Stats As Boolean)
        Dim Index As Integer
        SaveThingInfo()
        ThingsTree.SelectedNode = Nothing
        For Index = 8 To 1 Step -1
            Buffer(Index) = Buffer(Index - 1)
        Next
        Buffer(0) = New BufferItem
        With Buffer(0)
            .Thing = New Roton.Thing(NewThing)
            .Tile = NewTile
            .Stats = Stats
            SelectedColor = .Tile.Color
            SelectElement()
        End With
        LoadThingInfo()
        UpdateBuffer()
    End Sub

    Private Sub SetBoard(ByVal Index As Integer)
        Environment.SetBoardEditor(Index)
        If (Not (Board Is Nothing)) Then
            SaveBoardInfo()
        Else
            UpdateBoardLists()
        End If
        Board = Environment.CurrentBoard
        LoadBoardInfo()
        UpdateThingList()
        Thing = Board.Things(0)
        LoadThingInfo()
    End Sub

    Private Sub SaveBoardInfo()
        Dim Update As Boolean
        With Board
            If (.Name <> txtBoardName.Text) Then
                Update = True
            End If
            .Name = txtBoardName.Text
            .MaxShots = numBoardMaxShots.Value
            .IsDark = chkBoardDark.Checked
            .RestartZap = chkBoardRestart.Checked
            .ExitN = cmbBoardExitN.SelectedIndex
            .ExitS = cmbBoardExitS.SelectedIndex
            .ExitW = cmbBoardExitW.SelectedIndex
            .ExitE = cmbBoardExitE.SelectedIndex
        End With
        If (Update) Then
            UpdateBoardLists()
        End If
    End Sub

    Private Sub LoadBoardInfo()
        With Board
            txtBoardName.Text = .Name
            numBoardMaxShots.Value = .MaxShots
            chkBoardDark.Checked = .IsDark
            chkBoardRestart.Checked = .RestartZap
            cmbBoardExitN.SelectedIndex = .ExitN
            cmbBoardExitS.SelectedIndex = .ExitS
            cmbBoardExitW.SelectedIndex = .ExitW
            cmbBoardExitE.SelectedIndex = .ExitE
        End With
    End Sub

    Private Sub SaveWorldInfo()
        Dim Flags(15) As String

        With Environment.World
            .Ammo = numWorldAmmo.Value
            .Board = cmbWorldStart.SelectedIndex
            .EnergyCycles = numWorldEnergyCycles.Value
            Flags(0) = txtWorldFlag0.Text
            Flags(1) = txtWorldFlag1.Text
            Flags(2) = txtWorldFlag2.Text
            Flags(3) = txtWorldFlag3.Text
            Flags(4) = txtWorldFlag4.Text
            Flags(5) = txtWorldFlag5.Text
            Flags(6) = txtWorldFlag6.Text
            Flags(7) = txtWorldFlag7.Text
            Flags(8) = txtWorldFlag8.Text
            Flags(9) = txtWorldFlag9.Text
            Flags(10) = txtWorldFlagA.Text
            Flags(11) = txtWorldFlagB.Text
            Flags(12) = txtWorldFlagC.Text
            Flags(13) = txtWorldFlagD.Text
            Flags(14) = txtWorldFlagE.Text
            Flags(15) = txtWorldFlagF.Text
            ReDim Preserve Flags(Environment.Parameters.MaxFlags - 1)
            .Flags.Clear()
            .Flags.AddRange(Flags)
            .Gems = numWorldGems.Value
            .Health = numWorldHealth.Value
            .Keys(0) = chkWorldKeys0.Checked
            .Keys(1) = chkWorldKeys1.Checked
            .Keys(2) = chkWorldKeys2.Checked
            .Keys(3) = chkWorldKeys3.Checked
            .Keys(4) = chkWorldKeys4.Checked
            .Keys(5) = chkWorldKeys5.Checked
            .Keys(6) = chkWorldKeys6.Checked
            .Name = txtWorldName.Text
            .SavedGame = chkWorldSavedGame.Checked
            .Score = numWorldScore.Value
            .Stones = numWorldStones.Value
            .TimePassed = numWorldTimePassed.Value
            .TorchCycles = numWorldTorchCycles.Value
            .Torches = numWorldTorches.Value
        End With
    End Sub

    Private Sub LoadWorldInfo()
        Dim Flags(15) As String

        With Environment.World
            numWorldAmmo.Value = .Ammo
            cmbWorldStart.SelectedIndex = .Board
            numWorldEnergyCycles.Value = .EnergyCycles
            Flags = .Flags.ToArray
            ReDim Preserve Flags(15)
            txtWorldFlag0.Text = Flags(0)
            txtWorldFlag1.Text = Flags(1)
            txtWorldFlag2.Text = Flags(2)
            txtWorldFlag3.Text = Flags(3)
            txtWorldFlag4.Text = Flags(4)
            txtWorldFlag5.Text = Flags(5)
            txtWorldFlag6.Text = Flags(6)
            txtWorldFlag7.Text = Flags(7)
            txtWorldFlag8.Text = Flags(8)
            txtWorldFlag9.Text = Flags(9)
            txtWorldFlagA.Text = Flags(10)
            txtWorldFlagB.Text = Flags(11)
            txtWorldFlagC.Text = Flags(12)
            txtWorldFlagD.Text = Flags(13)
            txtWorldFlagE.Text = Flags(14)
            txtWorldFlagF.Text = Flags(15)
            numWorldGems.Value = .Gems
            numWorldHealth.Value = .Health
            chkWorldKeys0.Checked = .Keys(0)
            chkWorldKeys1.Checked = .Keys(1)
            chkWorldKeys2.Checked = .Keys(2)
            chkWorldKeys3.Checked = .Keys(3)
            chkWorldKeys4.Checked = .Keys(4)
            chkWorldKeys5.Checked = .Keys(5)
            chkWorldKeys6.Checked = .Keys(6)
            txtWorldName.Text = .Name
            chkWorldSavedGame.Checked = .SavedGame
            numWorldScore.Value = .Score
            numWorldStones.Value = .Stones
            numWorldTimePassed.Value = .TimePassed
            numWorldTorchCycles.Value = .TorchCycles
            numWorldTorches.Value = .Torches
        End With
    End Sub

    Private Sub SaveThingInfo()
        Dim Code As String = txtCode.Text.Replace(Chr(13) & Chr(10), Chr(13))
        With Thing
            .X = numThingX.Value
            .Y = numThingY.Value
            .XStep = numThingXStep.Value
            .YStep = numThingYStep.Value
            .Cycle = numThingCycle.Value
            .P1 = numThingP1.Value
            .P2 = numThingP2.Value
            .P3 = numThingP3.Value
            .Follower = numThingFollower.Value
            .Leader = numThingLeader.Value
            .CurrentInstruction = numThingInstruction.Value
            .UnderColor = numThingUnderColor.Value
            If (cmbThingUnderID.SelectedIndex >= 0) Then
                .UnderElement = Environment.Library.GetElementFromIndex(cmbThingUnderID.SelectedIndex)
            End If
            If (numThingBind.Value <> 0) Then
                .Length = -numThingBind.Value
                .Code = Board.Things(numThingBind.Value).Code
            ElseIf (Code <> "") Then
                If (.Code Is Nothing) OrElse (Code <> .Code.Code) Then
                    .Code = New Roton.CodeContainer(Code)
                End If
                .Length = Code.Length
            Else
                .Length = 0
                .Code = Nothing
            End If
            Environment.EditorUpdate(.X, .Y)
        End With
    End Sub

    Private Sub LoadThingInfo()
        Dim E As Roton.Element
        Dim Code As String
        With Thing
            numThingX.Value = .X
            numThingY.Value = .Y
            numThingXStep.Value = .XStep
            numThingYStep.Value = .YStep
            numThingCycle.Value = .Cycle
            numThingP1.Value = .P1
            numThingP2.Value = .P2
            numThingP3.Value = .P3
            numThingFollower.Value = .Follower
            numThingLeader.Value = .Leader
            numThingInstruction.Value = .CurrentInstruction
            numThingUnderColor.Value = .UnderColor
            cmbThingUnderID.SelectedIndex = Environment.Library.GetIndexFromElement(.UnderElement)
            If (.Length < 0) Then
                numThingBind.Value = -.Length
                Code = Board.Things(numThingBind.Value).Code.Code
            Else
                numThingBind.Value = 0
                If (Not (.Code Is Nothing)) Then
                    Code = .Code.Code
                Else
                    Code = ""
                End If
            End If
            If (Code Is Nothing) Then
                txtCode.Text = ""
            Else
                txtCode.Text = Code.Replace(Chr(13), Chr(13) & Chr(10))
            End If
            txtCode.Enabled = (.Length >= 0)
        End With
        If (ThingsTree.SelectedNode Is Nothing) Then
            grpThing.Text = "Buffer"
            E = Buffer(0).Tile.Element
        Else
            grpThing.Text = "Thing #" & ThingsTree.SelectedNode.Index
            E = Board.TileAt(Thing.X, Thing.Y).Element
        End If
        grpCode.Text = grpThing.Text
        If (E.P1 = "") Then
            btnThingP1.Text = "P1"
        Else
            btnThingP1.Text = E.P1
        End If
        If (E.P2 = "") Then
            btnThingP2.Text = "P2"
        Else
            btnThingP2.Text = E.P2
        End If
        If (E.P3 = "") Then
            btnThingP3.Text = "P3"
        Else
            btnThingP3.Text = E.P3
        End If
        If (E.Step = "") Then
            btnThingXStep.Text = "X Step"
            btnThingYStep.Text = "Y Step"
        Else
            btnThingXStep.Text = "X " & E.Step
            btnThingYStep.Text = "Y " & E.Step
        End If
        If (E.Board = "") Then
            P3IsBoard = False
        Else
            P3IsBoard = True
            btnThingP3.Text = E.Board
        End If
    End Sub

    Private Sub EditorForm_GotFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.GotFocus
        MainForm.ActiveEnvironment = Environment
    End Sub

    Private Sub ThingsTree_AfterSelect(ByVal sender As System.Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles ThingsTree.AfterSelect
        SelectThing(e.Node.Index)
    End Sub

    Private Sub ApplyAllChanges()
        If (Not (Thing Is Nothing)) Then
            SaveThingInfo()
        End If
        If (Not (Environment Is Nothing)) Then
            SaveWorldInfo()
            SaveBoardInfo()
        End If
    End Sub

    Private Sub BuildGame()
        Dim Exported As Roton.Environment
        Dim ExportStream As New IO.MemoryStream
        Dim Game As Form
        Roton.File.Save(ExportStream, Environment)
        ExportStream.Position = 0
        Exported = Roton.File.Load(ExportStream)
        Game = New GameForm(Exported, FontData)
        Game.MdiParent = MdiParent
        Game.Show()
    End Sub

    Private Sub TestGame()
        ApplyAllChanges()
        BuildGame()
    End Sub

    Private Sub TestBoard()
        Dim OldBoard As Integer
        Dim OldSaved As Boolean

        ApplyAllChanges()
        OldSaved = Environment.World.SavedGame
        OldBoard = Environment.World.Board
        Environment.World.SavedGame = True
        Environment.World.Board = BoardMenuIndex
        Environment.State.Execution = Roton.EnvironmentGameExecutionState.GameInit
        BuildGame()
        Environment.World.SavedGame = OldSaved
    End Sub

    Private Sub ButtonDown(ByVal sender As TSButton, ByVal e As System.Windows.Forms.MouseEventArgs)
        Dim X As Integer = e.X
        Dim Y As Integer = e.Y

        If (sender.DisplayStyle = ToolStripItemDisplayStyle.Image) Then
            X -= (sender.Width - sender.Image.Width) \ 2
            Y -= (sender.Height - sender.Image.Height) \ 2
        End If

        If (e.Button = Windows.Forms.MouseButtons.Left) Then
            Select Case sender.Name
                Case "Active"
                    SaveThingInfo()
                    ThingsTree.SelectedNode = Nothing
                    Thing = Buffer(0).Thing
                    LoadThingInfo()
                Case "Buffer"
                    SaveThingInfo()
                    ThingsTree.SelectedNode = Nothing
                    Buffer(0) = Buffer((X \ TileWidth) + 1)
                    Buffer(0).Thing = New Roton.Thing(Buffer(0).Thing)
                    SelectedColor = Buffer(0).Tile.Color
                    LoadThingInfo()
                    UpdateBuffer()
                Case "Stats"
                    Buffer(0).Stats = Not Buffer(0).Stats
                    StatsButton.Checked = Buffer(0).Stats
                Case "Default"
                    DefaultColorButton.Checked = Not DefaultColorButton.Checked
                    UpdateBuffer()
                Case "Hex"
                    mnuHexColor.Show(tsColors, HexButton.Bounds.Location)
            End Select
        End If
    End Sub

    Private Sub ColorDown(ByVal sender As TSButton, ByVal e As System.Windows.Forms.MouseEventArgs)
        If (e.Button = Windows.Forms.MouseButtons.Left) Then
            SelectedColor = (SelectedColor And &HF0) Or (CInt("&H" & sender.Text))
        ElseIf (e.Button = Windows.Forms.MouseButtons.Right) Then
            SelectedColor = (SelectedColor And &HF) Or (CInt("&H" & sender.Text) << 4)
        End If
        UpdateBuffer()
    End Sub

    Private Sub GrabTile(ByVal X As Integer, ByVal Y As Integer)
        AddBuffer(Board.ThingAt(X, Y), Board.TileAt(X, Y), Board.ThingIndexAt(X, Y) >= 0)
    End Sub

    Private Function PlotTile(ByVal X As Integer, ByVal Y As Integer)
        Dim T As Roton.Thing

        With Buffer(0)
            If (Not .Stats) OrElse (Board.Things.Count < Environment.Parameters.MaxThings) OrElse (Board.ThingIndexAt(X, Y) > 0) Then
                If (.Stats) Then
                    T = .Thing
                Else
                    T = Nothing
                End If
                Environment.EditorPlot(X, Y, .Tile, T, .Stats)
            End If
        End With

        Return (Board.TileAt(X, Y).Element.ID = Buffer(0).Tile.Element.ID)
    End Function

    Private Sub SelectThing(ByVal Index As Integer)
        If (Not (Thing Is Nothing)) Then
            SaveThingInfo()
        End If
        ThingsTree.SelectedNode = ThingsTree.Nodes(Index)
        Thing = Board.Things(Index)
        LoadThingInfo()
    End Sub

    Private Sub SelectElement()
        Dim TargetMenu As String
        With Buffer(0)
            StatsButton.Checked = (.Tile.Element.Cycle >= 0)
            .Thing.Cycle = .Tile.Element.Cycle
            .Thing.Follower = -1
            .Thing.Leader = -1
            .Thing.XStep = 0
            .Thing.YStep = 0
            .Stats = StatsButton.Checked
        End With
        UpdateBuffer()
        Thing = Buffer(0).Thing
        LoadThingInfo()
        TargetMenu = Buffer(0).Tile.Element.Menu.ToString
        If (TargetMenu = "5") Then
            TargetMenu = "3"
        End If
        ElementTree.SelectedNode = ElementTree.Nodes(TargetMenu).Nodes("E" & Environment.Library.GetIDFromElement(Buffer(0).Tile.Element.ID))
    End Sub

    Private Sub mnuRCPlayerMove_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuRCPlayerMove.Click
        Environment.EditorMovePlayer(MouseX, MouseY)
        If (Board.Things.IndexOf(Thing) = 0) Then
            LoadThingInfo()
        End If
        UpdateThingList()
    End Sub

    Private Sub mnuRCCopy_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuRCCopy.Click
        GrabTile(MouseX, MouseY)
    End Sub

    Private Sub mnuRCEdit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuRCEdit.Click
        Dim Index As Integer = Environment.EditorIndexAt(MouseX, MouseY)
        If (Index >= 0) Then
            SelectThing(Index)
            If (Thing.Code Is Nothing) OrElse (Thing.Length <= 0) Then
                tabMain2.SelectTab(tabThing)
            Else
                tabMain2.SelectTab(tabCode)
            End If
        End If
    End Sub

    Private Sub ElementTree_AfterSelect(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles ElementTree.AfterSelect
        Dim Index As Byte
        If (e.Node.Name.StartsWith("E")) Then
            ElementTree.SuspendLayout()
            Index = CByte(e.Node.Name.Substring(1))
            Buffer(0).Tile.Element = Environment.Library.GetElementFromIndex(Index)
            SelectElement()
            ElementTree.ResumeLayout()
            ElementTree.Update()
        End If
    End Sub

    Private Sub mnuBoardAdd_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuBoardAdd.Click
        Dim B As Roton.Board
        With Environment.Parameters
            B = New Roton.Board(.BoardWidth, .BoardHeight, .MaxThings, Environment.Library.GetElementFromID(Roton.ElementID.BoardEdge), Environment.Library.GetElementFromID(Roton.ElementID.Empty), Environment.Library.GetElementFromID(Roton.ElementID.Player))
            B.Clear()
        End With
        B.MaxShots = 255
        AddBoard(B)
    End Sub

    Private Sub AddBoard(ByVal B As Roton.Board)
        With Environment.Parameters
            If (Environment.Boards.Count < .MaxBoards) Then
                Environment.Boards.Add(B)
                Environment.World.Board = Environment.Boards.Count - 1
                UpdateBoardLists()
            End If
        End With
    End Sub

    Private Sub mnuBoardRemove_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuBoardRemove.Click
        If (BoardMenuIndex > 0) Then
            Environment.EditorRemoveBoard(BoardMenuIndex)
            If (Environment.World.Board > BoardMenuIndex) Then
                Environment.World.Board -= 1
            End If
            If (Environment.World.Board >= Environment.Boards.Count) Then
                Environment.World.Board = Environment.Boards.Count - 1
            End If
            SetBoard(Environment.World.Board)
            UpdateBoardLists()
        End If
    End Sub

    Private Sub btnThingApply_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnThingApply.Click
        SaveThingInfo()
        LoadThingInfo()
        tabMain2.SelectTab(tabEdit)
    End Sub

    Private Sub mnuBoardTest_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuBoardTest.Click
        TestBoard()
    End Sub

    Private Sub btnThingXStep_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnThingXStep.Click
        mnuThingStep.Show(btnThingXStep, 0, btnThingXStep.Height)
    End Sub

    Private Sub btnThingYStep_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnThingYStep.Click
        mnuThingStep.Show(btnThingYStep, 0, btnThingYStep.Height)
    End Sub

    Private Sub mnuThingStepNone_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuThingStepNone.Click
        numThingXStep.Value = 0
        numThingYStep.Value = 0
    End Sub

    Private Sub mnuThingStepNorth_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuThingStepNorth.Click
        numThingXStep.Value = 0
        numThingYStep.Value = -1
    End Sub

    Private Sub mnuThingStepSouth_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuThingStepSouth.Click
        numThingXStep.Value = 0
        numThingYStep.Value = 1
    End Sub

    Private Sub mnuThingStepWest_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuThingStepWest.Click
        numThingXStep.Value = -1
        numThingYStep.Value = 0
    End Sub

    Private Sub mnuThingStepEast_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuThingStepEast.Click
        numThingXStep.Value = 1
        numThingYStep.Value = 0
    End Sub

    Private Sub mnuCharList_ItemClicked(ByVal sender As Object, ByVal e As System.Windows.Forms.ToolStripItemClickedEventArgs) Handles mnuCharList.ItemClicked
        Dim NewChar As Integer = mnuCharList.Items.IndexOf(e.ClickedItem)
        numThingP1.Value = NewChar
        UpdateBuffer()
    End Sub

    Private Sub btnThingP1_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles btnThingP1.Click
        If (btnThingP1.Text.ToUpper = "CHARACTER?") Then
            mnuCharList.Show(btnThingP1, 0, btnThingP1.Height)
        End If
    End Sub

    Private Sub mnuHexColor_ItemClicked(ByVal sender As Object, ByVal e As System.Windows.Forms.ToolStripItemClickedEventArgs) Handles mnuHexColor.ItemClicked
        Dim NewColor As Integer = mnuHexColor.Items.IndexOf(e.ClickedItem)
        SelectedColor = NewColor
        UpdateBuffer()
    End Sub

    Private Sub mnuView2x_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuView2x.Click
        InitViewport()
    End Sub

    Private Sub btnWorldApply_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnWorldApply.Click
        SaveWorldInfo()
        LoadWorldInfo()
    End Sub

    Private Sub btnBoardApply_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnBoardApply.Click
        SaveBoardInfo()
        LoadBoardInfo()
    End Sub

    Private Sub tabMain2_TabIndexChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles tabMain2.TabIndexChanged
        If (tabMain2.SelectedIndex = 0) Then
            ApplyAllChanges()
        End If
    End Sub

    Private Sub mnuThingBind_ItemClicked(ByVal sender As Object, ByVal e As System.Windows.Forms.ToolStripItemClickedEventArgs) Handles mnuThingBind.ItemClicked
        numThingBind.Value = mnuThingBind.Items.IndexOf(e.ClickedItem)
        If (numThingBind.Value = 0) Then
            Thing.Code = New Roton.CodeContainer("")
        Else
            Thing.Code = Board.Things(numThingBind.Value).Code
            'txtCode.Text = Thing.Code.Code
        End If
        SaveThingInfo()
        LoadThingInfo()
    End Sub

    Private Sub btnThingBind_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnThingBind.Click
        mnuThingBind.Show(btnThingBind, 0, btnThingBind.Height)
    End Sub

    Private Sub mnuRCBind_DropDownItemClicked(ByVal sender As Object, ByVal e As System.Windows.Forms.ToolStripItemClickedEventArgs) Handles mnuRCBind.DropDownItemClicked
        Dim T As Roton.Thing = Board.ThingAt(MouseX, MouseY)
        T.Length = -mnuRCBind.DropDownItems.IndexOf(e.ClickedItem)
        LoadThingInfo()
    End Sub

    Private Sub BoardTree_NodeMouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeNodeMouseClickEventArgs) Handles BoardTree.NodeMouseClick
        BoardMenuIndex = e.Node.Index
        Select Case e.Button
            Case Windows.Forms.MouseButtons.Right
                mnuBoardName.Text = Environment.Boards(BoardMenuIndex).Name
                mnuBoard.Show(BoardTree, e.X, e.Y)
        End Select
    End Sub

    Private Function CreateObject(ByVal X As Integer, ByVal Y As Integer, ByVal C As Integer) As Integer
        Dim DesiredTile As Roton.Tile

        DesiredTile = New Roton.Tile(Environment.Library.GetElementFromID(Roton.ElementID.Object), C)
        Environment.EditorPlot(X, Y, DesiredTile, Nothing, True)
        Return Environment.EditorIndexAt(MouseX, MouseY)
    End Function

    Private Sub mnuRCObjectCreate_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuRCObjectCreate.Click
        Dim Index As Integer

        Index = CreateObject(MouseX, MouseY, SelectedColor)
        If (Index > 0) Then
            With Board.Things(Index)
                .P1 = 1
            End With
            UpdateThingList()
            SelectThing(Index)
            tabMain2.SelectTab(tabCode)
            Environment.EditorUpdate(MouseX, MouseY)
        End If
    End Sub

    Private Sub mnuRCObjectCreateHidden_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuRCObjectCreateHidden.Click
        Dim Index As Integer
        Dim Ch As Roton.AnsiChar

        Ch = Video.CharAt(MouseX, MouseY)
        Index = CreateObject(MouseX, MouseY, Ch.Color)
        If (Index > 0) Then
            With Board.Things(Index)
                .P1 = Ch.Char
            End With
            UpdateThingList()
            SelectThing(Index)
            tabMain2.SelectTab(tabCode)
            Environment.EditorUpdate(MouseX, MouseY)
        End If
    End Sub

    Private Sub BoardTree_AfterSelect(ByVal sender As System.Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles BoardTree.AfterSelect
        SetBoard(e.Node.Index)
    End Sub

    Private Sub mnuTopBoardStats_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuTopBoardStats.Click
        Dim TempStream As New IO.MemoryStream
        Dim Stats As String
        Roton.File.ExportBoard(TempStream, Board, Environment.Type)

        Stats = "File size: " & TempStream.Length & " bytes. (" & Math.Round(CDbl(TempStream.Length) / 1024, 1) & "k)"
        Stats &= vbCrLf & "Things: " & Board.Things.Count

        MsgBox(Stats, MsgBoxStyle.Information, "Board Statistics")
    End Sub

    Private Sub mnuTopWorldTest_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuTopWorldTest.Click
        TestGame()
    End Sub

    Private Sub mnuTopBoardTest_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuTopBoardTest.Click
        TestBoard()
    End Sub

    Private Sub mnuTopWorldStats_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuTopWorldStats.Click
        Dim TempStream As New IO.MemoryStream
        Dim Stats As String
        Roton.File.Save(TempStream, Environment)

        Stats = "File size: " & TempStream.Length & " bytes. (" & Math.Round(CDbl(TempStream.Length) / 1024, 1) & "k)"

        MsgBox(Stats, MsgBoxStyle.Information, "World Statistics")
    End Sub

    Private Sub mnuTopBoardClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuTopBoardClear.Click
        Board.Clear()
        Thing = Nothing
    End Sub

    Private Function OptimizeEmpties(ByVal B As Roton.Board) As Integer
        Dim Index As Integer = 0
        Dim Area As Integer = (B.Width * B.Height)
        Dim Ret As Integer = 0
        Do While (Index < Area)
            With B.Tiles(Index)
                If (.Element.ID = Roton.ElementID.Empty) Then
                    .Color = 0
                    Ret += 1
                End If
            End With
            Index += 1
        Loop
        Return Ret
    End Function

    Private Sub mnuOptimizeEmptiesBoard_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuOptimizeEmptiesBoard.Click
        Dim Ret As Integer = OptimizeEmpties(Board)
        MsgBox("Empties changed: " & Ret, MsgBoxStyle.Information, "Board Optimization")
    End Sub

    Private Sub mnuOptimizeEmptiesWorld_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuOptimizeEmptiesWorld.Click
        Dim Index As Integer
        Dim Ret As Integer = 0
        For Index = 0 To Environment.Boards.Count - 1
            Ret += OptimizeEmpties(Environment.Boards(Index))
        Next
        MsgBox("Empties changed: " & Ret, MsgBoxStyle.Information, "World Optimization")
    End Sub

    Private Sub mnuRCFill_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuRCFill.Click
        Dim FillCoords As New Queue(Of Coordinate)
        Dim FillTarget As Roton.Tile
        Dim TargetTile As Roton.Tile

        FillTarget = Board.TileAt(MouseX, MouseY)
        FillCoords.Enqueue(New Coordinate(MouseX, MouseY))

        Do Until (FillCoords.Count = 0)
            With FillCoords.Dequeue
                TargetTile = Board.TileAt(.X, .Y)
                If (TargetTile.Color = Buffer(0).Tile.Color) AndAlso (TargetTile.Element.ID = Buffer(0).Tile.Element.ID) Then
                    Continue Do
                End If
                If (Not PlotTile(.X, .Y)) Then
                    Continue Do
                End If
                TargetTile = Board.TileAt(.X - 1, .Y)
                If (TargetTile.Color = FillTarget.Color) AndAlso (TargetTile.Element.ID = FillTarget.Element.ID) Then
                    FillCoords.Enqueue(New Coordinate(.X - 1, .Y))
                End If
                TargetTile = Board.TileAt(.X, .Y - 1)
                If (TargetTile.Color = FillTarget.Color) AndAlso (TargetTile.Element.ID = FillTarget.Element.ID) Then
                    FillCoords.Enqueue(New Coordinate(.X, .Y - 1))
                End If
                TargetTile = Board.TileAt(.X + 1, .Y)
                If (TargetTile.Color = FillTarget.Color) AndAlso (TargetTile.Element.ID = FillTarget.Element.ID) Then
                    FillCoords.Enqueue(New Coordinate(.X + 1, .Y))
                End If
                TargetTile = Board.TileAt(.X, .Y + 1)
                If (TargetTile.Color = FillTarget.Color) AndAlso (TargetTile.Element.ID = FillTarget.Element.ID) Then
                    FillCoords.Enqueue(New Coordinate(.X, .Y + 1))
                End If
            End With
        Loop
        UpdateThingList()
    End Sub

    Private Sub ExportBoard(ByVal B As Roton.Board)
        Dim TempStream As IO.FileStream
        With SFDBoard
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                TempStream = New IO.FileStream(.FileName, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
                Roton.File.ExportBoard(TempStream, Board, Environment.Type)
                TempStream.Close()
            End If
        End With
    End Sub

    Private Function ImportBoard() As Roton.Board
        Dim TempStream As IO.FileStream
        Dim Ret As Roton.Board = Nothing

        With OFDBoard
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                TempStream = New IO.FileStream(.FileName, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
                Ret = Roton.File.ImportBoard(TempStream, Environment.Type)
                TempStream.Close()
            End If
        End With
        Return Ret
    End Function

    Private Sub mnuBoardTopStart_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuBoardTopStart.Click
        cmbWorldStart.SelectedIndex = Environment.Boards.IndexOf(Board)
    End Sub

    Private Sub mnuBoardExport_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuBoardExport.Click, mnuBoardTopExport.Click
        ExportBoard(Board)
    End Sub

    Private Sub mnuBoardImport_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuBoardImport.Click, mnuBoardTopImport.Click
        Dim B As Roton.Board
        B = ImportBoard()
        If (Not (B Is Nothing)) Then
            AddBoard(B)
        End If
    End Sub

    Private Sub mnuTopBoardScreenshot_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuTopBoardScreenshot.Click
        Dim G As Graphics
        Dim B As Bitmap

        With SFDScreenshot
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                B = New Bitmap(TileWidth * Board.Width, TileHeight * Board.Height, Imaging.PixelFormat.Format32bppPArgb)
                G = Graphics.FromImage(B)
                G.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
                G.PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
                G.DrawImage(Video.Bitmap, 0, 0, B.Width, B.Height)
                B.Save(.FileName, Imaging.ImageFormat.Png)
            End If
        End With
    End Sub
End Class