﻿Public Class GameForm
    Public Environment As Roton.Environment
    Private Video As Roton.VideoGDI
    Private HUDVideo As Roton.VideoGDI
    Private Buffer As Bitmap
    Private Input As Roton.InputForm
    Private LastBoard As Integer = -1
    Private XScale As Integer
    Private YScale As Integer
    Private XZoom As Integer = 1
    Private YZoom As Integer = 1
    Private XCamera As Integer
    Private YCamera As Integer
    Private HudWidth As Integer
    Private HudHeight As Integer
    Private Status As ToolStripStatusLabel
    Private CameraBoundWidth As Integer
    Private CameraBoundHeight As Integer
    Private Resized As Boolean
    Private Cheat As String = ""
    Private Cheating As Boolean
    Private WithEvents GameTimer As Roton.Timer

    Public Sub New(ByVal NewEnvironment As Roton.Environment, ByVal Font() As Byte)
        ' This call is required by the designer.
        InitializeComponent()

        ' Add any initialization after the InitializeComponent() call.
        Environment = NewEnvironment

        ' set up video
        Video = New Roton.VideoGDI
        Video.SetFont(Font)
        Video.SetViewport(Environment.Parameters.ViewWidth, Environment.Parameters.ViewHeight)

        HudWidth = 20
        HudHeight = Environment.Parameters.ViewHeight
        HUDVideo = New Roton.VideoGDI
        HUDVideo.SetFont(Font)
        HUDVideo.SetSize(HudWidth, HudHeight)
        HUDVideo.SetViewport(HudWidth, HudHeight)

        Environment.Video = Video
        If Environment.Parameters.DisplayWide Then
            XScale = 2
            YScale = 1
        Else
            XScale = 1
            YScale = 1
        End If
        Buffer = New Bitmap((Video.Bitmap.Width * XScale) + HUDVideo.Bitmap.Width, Video.Bitmap.Height, Imaging.PixelFormat.Format32bppPArgb)
        UpdateSize()
        BackgroundImage = Buffer 'Video.Bitmap

        ' set up input
        Input = New Roton.InputForm
        Environment.Input = Input

        ' set up messenger
        Environment.Messenger = New Roton.MessengerVideo(Video, Environment.Parameters)

        ' set up scroll
        Environment.Scroll = New Roton.ScrollVideo(Video, Environment.State, Environment.Parameters)

        ' set up camera stuff
        CameraBoundWidth = Video.ViewWidth \ 3
        CameraBoundHeight = Video.ViewHeight \ 3

        ' set gamespeed
        SetGameSpeed(4)
    End Sub

    Private Function HudNumber(ByVal InNumber As Integer) As String
        Dim Ret As String = InNumber & " "
        If (InNumber >= 0) Then
            Ret = " " & Ret
        End If
        Return Ret
    End Function

    Private Sub UpdateHud()
        Dim Index As Integer
        Dim StoneString As String = ""

        Select Case Environment.State.PlayerElement.ID
            Case Roton.ElementID.Player
                ' special counter
                If (Environment.World.Stones >= 0) Then
                    For Index = 0 To (Environment.World.Flags.Count - 1)
                        If (Environment.World.Flags(Index).StartsWith("Z")) Then
                            StoneString = Environment.World.Flags(Index).Substring(1) & ":"
                        End If
                    Next
                    HUDVideo.WriteOverlay(12 - StoneString.Length, 12, StoneString & Environment.World.Stones & " ", &H1E)
                Else
                    HUDVideo.WriteOverlay(2, 12, "                 ", &H1E)
                End If

                ' time
                If (Environment.CurrentBoard.TimeLimit > 0) Then
                    HUDVideo.WriteOverlay(7, 5, "Time:" & Environment.CurrentBoard.TimeLimit - Environment.World.TimePassed & " ", &H1E)
                End If

                ' keys
                For Index = 0 To 6
                    If (Environment.World.Keys(Index)) Then
                        HUDVideo.PlotOverlay(12 + Index, 11, New Roton.AnsiChar(12, &H10 Or (Index + 9)))
                    Else
                        HUDVideo.PlotOverlay(12 + Index, 11, New Roton.AnsiChar(32, &H10))
                    End If
                Next

                HUDVideo.WriteOverlay(12, 6, Environment.World.Health & " ", &H1E)
                HUDVideo.WriteOverlay(12, 7, Environment.World.Ammo & " ", &H1E)
                HUDVideo.WriteOverlay(12, 8, Environment.World.Torches & " ", &H1E)
                HUDVideo.WriteOverlay(12, 9, Environment.World.Gems & " ", &H1E)
                HUDVideo.WriteOverlay(12, 10, Environment.World.Score & " ", &H1E)
            Case Roton.ElementID.Monitor
                HUDVideo.WriteOverlay(9, 8, Environment.World.Name, &H1F)
                HUDVideo.WriteOverlay(6, 22, "         ", &H1F)
                HUDVideo.PlotOverlay(7 + Environment.State.GameSpeed, 22, New Roton.AnsiChar(31, &H1F))
        End Select

        '' player info
        'HUDVideo.WriteOverlay(1, 3, "(" & Environment.PlayerX & ", " & Environment.PlayerY & ")    ", &H1C)

        If (Environment.State.GamePaused) Then
            HUDVideo.WriteOverlay(4, 4, "Pausing...", &H1F)
        End If

        HUDVideo.Update()
    End Sub

    Private Sub CheatHud()
        HUDVideo.WriteOverlay(2, 3, "            ", &HF)
        HUDVideo.WriteOverlay(2, 3, Cheat & "_", &HF)
    End Sub

    Private Sub RedrawHud()
        Dim X As Integer
        Dim Y As Integer

        For X = 0 To HudWidth - 1
            For Y = 0 To HudHeight - 1
                HUDVideo.PlotOverlay(X, Y, New Roton.AnsiChar(32, &H1E))
            Next
        Next

        For X = 0 To 6 Step 2
            HUDVideo.PlotOverlay(9 - X, 0, New Roton.AnsiChar(196, &H1F))
            HUDVideo.PlotOverlay(10 + X, 0, New Roton.AnsiChar(196, &H1F))
            HUDVideo.PlotOverlay(9 - X, 2, New Roton.AnsiChar(196, &H1F))
            HUDVideo.PlotOverlay(10 + X, 2, New Roton.AnsiChar(196, &H1F))
        Next

        HUDVideo.WriteOverlay(2, 1, "      Lyon      ", &H70)

        Select Case Environment.State.PlayerElement.ID
            Case Roton.ElementID.Player
                ' text and values
                HUDVideo.WriteOverlay(5, 6, "Health:", &H1E)
                HUDVideo.WriteOverlay(7, 7, "Ammo:", &H1E)
                HUDVideo.WriteOverlay(4, 8, "Torches:", &H1E)
                HUDVideo.WriteOverlay(7, 9, "Gems:", &H1E)
                HUDVideo.WriteOverlay(6, 10, "Score:", &H1E)
                HUDVideo.WriteOverlay(7, 11, "Keys:", &H1E)

                HUDVideo.WriteOverlay(2, 15, " B ", &H30)
                HUDVideo.WriteOverlay(7, 18, " " & Chr(24) & Chr(25) & Chr(26) & Chr(27), &H30)
                HUDVideo.WriteOverlay(1, 19, " Shift " & Chr(24) & Chr(25) & Chr(26) & Chr(27), &H70)
                HUDVideo.WriteOverlay(2, 21, " - ", &H70)
                HUDVideo.WriteOverlay(2, 22, " P ", &H30)
                HUDVideo.WriteOverlay(2, 23, " Q ", &H70)

                If (Environment.Parameters.TorchesEnabled) Then
                    HUDVideo.WriteOverlay(2, 14, " T ", &H70)
                    HUDVideo.WriteOverlay(6, 14, "Torch", &H1F)
                End If

                HUDVideo.WriteOverlay(6, 15, "Be quiet", &H1F)
                If (Environment.Parameters.HintsEnabled) Then
                    HUDVideo.WriteOverlay(6, 16, "Hint", &H1F)
                    HUDVideo.WriteOverlay(2, 16, " H ", &H70)
                Else
                    HUDVideo.WriteOverlay(6, 16, "Help", &H18)
                    HUDVideo.WriteOverlay(2, 16, " - ", &H70)
                End If
                HUDVideo.WriteOverlay(13, 18, "Move", &H1F)
                HUDVideo.WriteOverlay(13, 19, "Shoot", &H1F)
                HUDVideo.WriteOverlay(6, 21, "Save game", &H18)
                HUDVideo.WriteOverlay(6, 22, "Pause", &H1F)
                HUDVideo.WriteOverlay(6, 23, "Quit", &H1F)

                HUDVideo.PlotOverlay(2, 6, New Roton.AnsiChar(2, &H1F))
                HUDVideo.PlotOverlay(2, 7, New Roton.AnsiChar(132, &H1B))
                HUDVideo.PlotOverlay(2, 8, New Roton.AnsiChar(157, &H16))
                HUDVideo.PlotOverlay(2, 9, New Roton.AnsiChar(4, &H1B))
                HUDVideo.PlotOverlay(2, 11, New Roton.AnsiChar(12, &H1F))
            Case Roton.ElementID.Monitor
                HUDVideo.WriteOverlay(2, 5, "Pick a command:", &H1B)

                HUDVideo.WriteOverlay(2, 7, " - ", &H30)
                HUDVideo.WriteOverlay(2, 11, " P ", &H70)
                HUDVideo.WriteOverlay(2, 12, " - ", &H30)
                HUDVideo.WriteOverlay(2, 13, " Q ", &H70)
                HUDVideo.WriteOverlay(2, 16, " A ", &H30)
                HUDVideo.WriteOverlay(2, 17, " H ", &H70)
                HUDVideo.WriteOverlay(2, 18, " - ", &H30)
                HUDVideo.WriteOverlay(2, 21, " - ", &H70)

                HUDVideo.WriteOverlay(6, 7, "World:", &H1E)
                HUDVideo.WriteOverlay(6, 11, "Play", &H1F)
                HUDVideo.WriteOverlay(6, 12, "Restore game", &H18)
                HUDVideo.WriteOverlay(6, 13, "Quit", &H1E)
                HUDVideo.WriteOverlay(6, 16, "About", &H18)
                HUDVideo.WriteOverlay(6, 17, "High Scores", &H18)
                HUDVideo.WriteOverlay(6, 18, "Board Editor", &H18)
                HUDVideo.WriteOverlay(6, 21, "Game speed:", &H1E)
                HUDVideo.WriteOverlay(6, 23, "F....:....S", &H1E)
        End Select
    End Sub

    Private Sub UpdateSize()
        If (WindowState = FormWindowState.Normal) Then
            With Buffer
                ClientSize = New Size(.Width * XZoom, .Height * YZoom)
            End With
        End If
    End Sub

    Protected Overrides Sub OnPaintBackground(ByVal e As System.Windows.Forms.PaintEventArgs)
        Dim DisplayWidth As Integer
        Dim DisplayHeight As Integer
        Dim DisplayRatio As Double = ClientSize.Width / ClientSize.Height
        Dim ImageRatio As Double = (BackgroundImage.Width) / (BackgroundImage.Height)
        Dim DisplayX As Integer
        Dim DisplayY As Integer

        If (DisplayRatio > ImageRatio) Then
            DisplayHeight = ClientSize.Height
            DisplayWidth = DisplayHeight * ImageRatio
        Else
            DisplayWidth = ClientSize.Width
            DisplayHeight = DisplayWidth / ImageRatio
        End If

        DisplayX = (ClientSize.Width - DisplayWidth) \ 2
        DisplayY = (ClientSize.Height - DisplayHeight) \ 2

        With e.Graphics
            If (mnuViewInterpolated.Checked) AndAlso ((ClientSize.Width <> DisplayWidth) OrElse (ClientSize.Height <> DisplayHeight)) Then
                .InterpolationMode = Drawing2D.InterpolationMode.Low
            Else
                .InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
            End If
            .PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
            .DrawImage(BackgroundImage, DisplayX, DisplayY, DisplayWidth, DisplayHeight)
        End With
    End Sub

    Protected Sub SaveScreenshot()
        Dim SFD As New SaveFileDialog
        With SFD
            .Filter = "PNG screenshots|*.png"
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                BackgroundImage.Save(.FileName, Imaging.ImageFormat.Png)
            End If
        End With
    End Sub

    Protected Overrides Function ProcessCmdKey(ByRef msg As System.Windows.Forms.Message, ByVal keyData As System.Windows.Forms.Keys) As Boolean
        Dim EventArgs As KeyEventArgs = New KeyEventArgs(keyData)
        Input.KeyDown(EventArgs)
        Select Case EventArgs.KeyCode
            Case Keys.Up, Keys.Down, Keys.Left, Keys.Right,
                Keys.ShiftKey, Keys.LShiftKey, Keys.RShiftKey, Keys.Tab
                'these keys should always be processed by the game window
                Return True
        End Select
        Return MyBase.ProcessCmdKey(msg, keyData)
    End Function

    Private Sub GameForm_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.DoubleClick
        SaveScreenshot()
    End Sub

    Private Sub GameForm_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        Environment = Nothing
        Video = Nothing
    End Sub

    Private Sub GameForm_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        MainForm.ActiveEnvironment = Nothing
    End Sub

    Private Sub GameForm_GotFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.GotFocus
        MainForm.ActiveEnvironment = Environment
    End Sub

    Private Sub GameForm_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        If (Not Cheating) Then
            Input.KeyDown(e)
        End If
    End Sub

    Private Sub GameForm_KeyPress(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyPressEventArgs) Handles Me.KeyPress
        If (Not Cheating) Then
            If (e.KeyChar = "?") Then
                Cheating = True
                CheatHud()
                HUDVideo.Update()
                DrawBuffer()
                Invalidate()
            End If
        Else
            Select Case e.KeyChar
                Case Chr(8)
                    'backspace
                    If (Cheat.Length > 0) Then
                        Cheat = Cheat.Substring(0, Cheat.Length - 1)
                    End If
                Case Chr(13)
                    'enter
                    Cheating = False
                    RedrawHud()
                    UpdateHud()
                    Environment.Cheat(Cheat)
                    Cheat = ""
                Case Chr(27)
                    'escape
                    Cheating = False
                    RedrawHud()
                    UpdateHud()
                    Cheat = ""
                Case Else
                    Cheat &= e.KeyChar
            End Select
            If (Cheat.Length > 11) Then
                Cheat = Cheat.Substring(0, 11)
            End If
            If (Cheating) Then
                CheatHud()
            End If
            HUDVideo.Update()
            DrawBuffer()
            Invalidate()
        End If
    End Sub

    Private Sub GameForm_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        Input.KeyUp(e)
    End Sub

    Private Sub UpdateCamera()
        Dim CameraBoundLeft As Integer
        Dim CameraBoundRight As Integer
        Dim CameraBoundTop As Integer
        Dim CameraBoundBottom As Integer

        CameraBoundLeft = XCamera + CameraBoundWidth
        CameraBoundTop = YCamera + CameraBoundHeight
        CameraBoundRight = ((XCamera + Video.ViewWidth) - CameraBoundWidth) - 1
        CameraBoundBottom = ((YCamera + Video.ViewHeight) - CameraBoundHeight) - 1

        If (Environment.PlayerX < CameraBoundLeft) Then
            XCamera = Environment.PlayerX - CameraBoundWidth
        End If
        If (Environment.PlayerX > CameraBoundRight) Then
            XCamera = ((Environment.PlayerX - Video.ViewWidth) + CameraBoundWidth) + 1
        End If
        If (Environment.PlayerY < CameraBoundTop) Then
            YCamera = Environment.PlayerY - CameraBoundHeight
        End If
        If (Environment.PlayerY > CameraBoundBottom) Then
            YCamera = ((Environment.PlayerY - Video.ViewHeight) + CameraBoundHeight) + 1
        End If

        Video.SetCamera(XCamera, YCamera)
    End Sub

    Private Sub GameTimerTick(ByVal e As System.EventArgs)
        Tick()
    End Sub

    Private Sub DrawBuffer()
        Dim G As Graphics = Graphics.FromImage(Buffer)
        G.InterpolationMode = Drawing2D.InterpolationMode.NearestNeighbor
        G.PixelOffsetMode = Drawing2D.PixelOffsetMode.Half
        G.DrawImage(Video.Bitmap, 0, 0, Video.Bitmap.Width * XScale, Video.Bitmap.Height * YScale)
        G.DrawImageUnscaled(HUDVideo.Bitmap, Video.Bitmap.Width * XScale, 0)
    End Sub

    Private Sub Tick()
        Dim Updated As Boolean

        If (Not Focused) AndAlso (Not mnuOptionsBackground.Checked) Then
            Exit Sub
        End If

        If (Not Cheating) AndAlso (Not (Environment Is Nothing)) Then
            Updated = Environment.Execute
            If (Environment.State.RedrawHUD) Then
                Environment.State.RedrawHUD = False
                Environment.State.UpdateHUD = False
                RedrawHud()
                UpdateHud()
            End If
            If (Environment.State.UpdateHUD) Then
                Environment.State.UpdateHUD = False
                UpdateHud()
            End If
            If (Updated) Then
                UpdateCamera()
                Video.Update()
                DrawBuffer()
                Invalidate(ClientRectangle)
            End If
        End If
        If (Environment Is Nothing) OrElse (Environment.State.QuitZZT) Then
            Close()
        ElseIf (LastBoard <> Environment.World.Board) Then
            SetTitle()
            XCamera = Environment.PlayerX - (Video.ViewWidth \ 2)
            YCamera = Environment.PlayerY - (Video.ViewHeight \ 2)
            LastBoard = Environment.World.Board
        End If
    End Sub

    Private Sub SetTitle()
        Text = Environment.World.Name & " - " & Environment.CurrentBoard.Name
    End Sub

    Private Sub SetGameSpeed(ByVal Index As Integer)
        Environment.State.GameSpeed = Index
        Dim ItemIndex As Integer
        Dim X As ToolStripMenuItem

        For ItemIndex = 0 To mnuSpeed.DropDownItems.Count - 1
            X = mnuSpeed.DropDownItems(ItemIndex)
            X.Checked = (ItemIndex = Index)
        Next
    End Sub

    Private Sub mnuSpeed0_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed0.Click
        SetGameSpeed(0)
    End Sub

    Private Sub mnuSpeed1_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed1.Click
        SetGameSpeed(1)
    End Sub

    Private Sub mnuSpeed2_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed2.Click
        SetGameSpeed(2)
    End Sub

    Private Sub mnuSpeed3_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed3.Click
        SetGameSpeed(3)
    End Sub

    Private Sub mnuSpeed4_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed4.Click
        SetGameSpeed(4)
    End Sub

    Private Sub mnuSpeed5_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed5.Click
        SetGameSpeed(5)
    End Sub

    Private Sub mnuSpeed6_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed6.Click
        SetGameSpeed(6)
    End Sub

    Private Sub mnuSpeed7_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed7.Click
        SetGameSpeed(7)
    End Sub

    Private Sub mnuSpeed8_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuSpeed8.Click
        SetGameSpeed(8)
    End Sub

    Private Sub mnuView2x_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuView2x.Click
        If (mnuView2x.Checked) Then
            XZoom = 2
            YZoom = 2
        Else
            XZoom = 1
            YZoom = 1
        End If
        UpdateSize()
    End Sub

    Private Sub mnuViewResetSize_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewResetSize.Click
        UpdateSize()
        Resized = False
    End Sub

    Private Sub GameForm_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        mnuViewResetSize.Enabled = (WindowState = FormWindowState.Normal)
        mnuView2x.Enabled = (WindowState = FormWindowState.Normal)
    End Sub

    Private Sub GameForm_Shown(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Shown
        GameTimer = New Roton.Timer(1000000.0# / 71.2)
        AddHandler GameTimer.Tick, AddressOf GameTimerTick
        tmrTick.Enabled = True
    End Sub

    Private Sub tmrTick_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tmrTick.Tick
        GameTimer.Proc()
    End Sub
End Class
