﻿Public Class MainForm
    Private _ActiveEnvironment As Roton.Environment
    Private _DefaultFont() As Byte = My.Resources.vga_rom

    Public Property ActiveEnvironment As Roton.Environment
        Get
            Return _ActiveEnvironment
        End Get
        Set(ByVal value As Roton.Environment)
            _ActiveEnvironment = value
            mnuFileSave.Enabled = Not (_ActiveEnvironment Is Nothing)
        End Set
    End Property

    Private Sub mnuFileOpen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileOpen.Click
        Dim Env As Roton.Environment = LoadWorld()

        If (Not (Env Is Nothing)) Then
            StartGame(Env, OFDWorld.FileName).Show()
        End If
    End Sub

    Private Sub mnuFileExit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileExit.Click
        Application.Exit()
    End Sub

    Private Sub MainForm_DragDrop(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles Me.DragDrop
        Dim Files() As String = e.Data.GetData("FileDrop")
        Dim Env As New List(Of Roton.Environment)
        Dim FS As IO.Stream
        Dim KS As Integer = e.KeyState
        Dim Index As Integer
        Dim Forms As New List(Of Form)

        ' ctrl=8
        ' alt=32
        ' shift=4

        For Each File In Files
            FS = New IO.FileStream(File, IO.FileMode.Open, IO.FileAccess.Read)
            Env.Add(Roton.File.Load(FS))
            FS.Close()
        Next

        ' if ctrl is pressed, open in editor
        If ((e.KeyState And 8) <> 0) Then
            For Index = 0 To Env.Count - 1
                If (Not (Env(Index) Is Nothing)) Then
                    StartEditor(Env(Index), Files(Index)).Show()
                End If
            Next
        Else
            For Index = 0 To Env.Count - 1
                If (Not (Env(Index) Is Nothing)) Then
                    StartGame(Env(Index), Files(Index)).Show()
                End If
            Next
        End If
    End Sub

    Private Sub MainForm_DragOver(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles Me.DragOver
        e.Effect = DragDropEffects.All
    End Sub

    Private Sub MainForm_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        ActiveEnvironment = Nothing
    End Sub

    Private Sub mnuFileSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileSave.Click
        Dim FileName As String
        Dim FileStream As IO.FileStream

        With _ActiveEnvironment
            If (Not .State.EditorMode) AndAlso (Not .State.TitleScreen) Then
                .State.GamePaused = True
            End If
            If (.World.SavedGame) Then
                SFDWorld.Filter = "Saved games|*.sav|All files|*.*"
                SFDWorld.DefaultExt = "sav"
            Else
                Select Case (.Type)
                    Case Roton.EnvironmentGameType.ZZT
                        SFDWorld.Filter = "ZZT worlds|*.zzt|All files|*.*"
                        SFDWorld.DefaultExt = "zzt"
                    Case Roton.EnvironmentGameType.SuperZZT
                        SFDWorld.Filter = "Super ZZT worlds|*.szt|All files|*.*"
                        SFDWorld.DefaultExt = "szt"
                    Case Else
                        SFDWorld.Filter = "All files|*.*"
                        SFDWorld.DefaultExt = ""
                End Select
            End If
        End With

        With SFDWorld
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                FileName = .FileName
                FileStream = New IO.FileStream(FileName, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
                Roton.File.Save(FileStream, _ActiveEnvironment)
                FileStream.Close()
            End If
        End With
    End Sub

    Private Function GetFont(ByVal FileName As String) As Byte()
        Dim Ret() As Byte = _DefaultFont
        Dim F As String
        Dim D As String
        Dim B() As Byte

        With My.Computer.FileSystem
            F = .GetName(FileName)
            D = FileName.Substring(0, FileName.Length - F.Length)
            If (F.Contains(".")) Then
                F = F.Substring(0, F.IndexOf("."))
            End If
            D &= F
            If (.FileExists(D & ".com")) Then
                B = .ReadAllBytes(D & ".com")
                If (B.Length > 3584) AndAlso (B(0) = &HEB) AndAlso (B(1) = &H47) AndAlso (B(2) = &HA2) AndAlso (B(3) = &H5) Then
                    'font mania 2.2
                    ReDim Ret(B.Length - &H5A4)
                    Array.ConstrainedCopy(B, &H5A2, Ret, 0, Ret.Length)
                End If
            ElseIf (.FileExists(D & ".chr")) Then
                B = .ReadAllBytes(D & ".chr")
                If (B.Length = 3584) Then
                    'megazeux chr
                    ReDim Ret(3583)
                    Array.Copy(B, Ret, 3584)
                End If
            End If
        End With

        Return Ret
    End Function

    Private Function LoadWorld() As Roton.Environment
        Dim FileName As String
        Dim FileStream As IO.FileStream
        Dim Env As Roton.Environment = Nothing

        With OFDWorld
            If (.ShowDialog = Windows.Forms.DialogResult.OK) Then
                FileName = .FileName
                FileStream = New IO.FileStream(FileName, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
                Env = Roton.File.Load(FileStream)
                FileStream.Close()
            End If
        End With

        Return Env
    End Function

    Private Function StartGame(ByVal Env As Roton.Environment, ByVal FileName As String) As Form
        Dim Game As Form

        Game = New GameForm(Env, GetFont(FileName))
        Game.MdiParent = Me
        Return Game
    End Function

    Private Function StartEditor(ByVal Env As Roton.Environment, ByVal FileName As String) As Form
        Dim Editor As Form

        Editor = New EditorForm(Env, GetFont(FileName), lblStatus)
        Editor.MdiParent = Me
        Return Editor
    End Function

    Private Sub mnuFileEdit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileEdit.Click
        Dim Env As Roton.Environment = LoadWorld()

        If (Not (Env Is Nothing)) Then
            StartEditor(Env, OFDWorld.FileName).Show()
        End If
    End Sub

    Private Sub mnuFileNewZZT_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileNewZZT.Click
        StartEditor(Roton.File.Create(Roton.EnvironmentGameType.ZZT), "").Show()
    End Sub

    Private Sub mnuFileNewSuperZZT_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileNewSuperZZT.Click
        StartEditor(Roton.File.Create(Roton.EnvironmentGameType.SuperZZT), "").Show()
    End Sub

    Private Sub mnuArrangeTileH_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuArrangeTileH.Click
        LayoutMdi(MdiLayout.TileHorizontal)
    End Sub

    Private Sub mnuArrangeTileV_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuArrangeTileV.Click
        LayoutMdi(MdiLayout.TileVertical)
    End Sub

    Private Sub mnuArrangeCascade_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuArrangeCascade.Click
        LayoutMdi(MdiLayout.Cascade)
    End Sub
End Class