﻿Imports Microsoft.DirectX.DirectSound
Imports System.IO

Namespace Roton
    Public Class AudioDirectX9
        Implements Audio

        Private Dev As Device
        Private Sounds(9) As Buffer
        Private Tone As Buffer
        Private Played As Buffer

        Public Sub New(ByVal OwnerForm As Form)
            Dev = New Device
            Dev.SetCooperativeLevel(OwnerForm, CooperativeLevel.Normal)
        End Sub

        Protected Overrides Sub Finalize()
            MyBase.Finalize()
            Dev.Dispose()
            Dev = Nothing
        End Sub

        Public Sub PlaySound(ByVal SoundIndex As Integer) Implements Audio.PlaySound
            If Not (Played Is Nothing) Then
                Played.Stop()
            End If
            Played = Sounds(SoundIndex)
            Played.Play(0, BufferPlayFlags.TerminateByTime)
        End Sub

        Public Sub PlayTone(ByVal Tone As Integer) Implements Audio.PlayTone
            Dim Mem As New MemoryStream
            Dim Writer As New BinaryWriter(Mem)
            Dim I32 As Int32
            Dim I16 As Int16
            Dim HighSample As Byte = &HC0
            Dim LowSample As Byte = &H40
            Dim Freq As Int32 = 44100
            Dim Octave As Integer = (Tone >> 4)
            Dim Note As Integer = (Tone And &HF) + (Octave * 12)
            Dim Length As Integer = Freq / (440 * (2 ^ ((Note - 40) / 12)))
            Dim PeakLength As Integer = Length >> 1
            Dim Wave(Length) As Byte
            Dim Index As Integer

            For Index = 0 To PeakLength - 1
                Wave(Index) = HighSample
                Wave(Index + PeakLength) = LowSample
            Next

            I32 = &H46464952
            Writer.Write(I32)   '"RIFF"
            I32 = Length + 36
            Writer.Write(I32)   'RIFF length
            I32 = &H45564157
            Writer.Write(I32)   '"WAVE"
            I32 = &H20746D66
            Writer.Write(I32)   '"fmt "
            I32 = &H10
            Writer.Write(I32)   'format chunk length
            I16 = 1
            Writer.Write(I16)   'format
            I16 = 1
            Writer.Write(I16)   'channels
            Writer.Write(Freq)  'samples / sec
            Writer.Write(Freq)  'bytes / sec
            I16 = 1
            Writer.Write(I16)   'bytes / sample
            I16 = 8
            Writer.Write(I16)   'bits / sample
            I32 = &H61746164
            Writer.Write(I32)   '"data"
            I32 = Length
            Writer.Write(I32)   'data length
            Writer.Write(Wave)
        End Sub

        Public WriteOnly Property SoundEffect(ByVal Index As Integer) As Byte() Implements Audio.SoundEffect
            Set(ByVal value As Byte())
                Dim Mem As New MemoryStream(Index)
                Sounds(Index) = New Buffer(Mem, Dev)
            End Set
        End Property

        Public Sub [Stop]() Implements Audio.Stop
            If Not (Played Is Nothing) Then
                Played.Stop()
            End If
        End Sub
    End Class
End Namespace