﻿Namespace Roton
    Partial Public Class Environment

        Public Enum OOPCodeSets As Integer
            Normal = 0
            BQuest = 1
        End Enum

        Protected Shared ReadOnly OOPCodeSetNormal() As String = {
            "THEN", "GO", "TRY", "WALK", "SET",
            "CLEAR", "IF", "SHOOT", "THROWSTAR", "GIVE",
            "TAKE", "END", "ENDGAME", "IDLE", "RESTART",
            "ZAP", "RESTORE", "LOCK", "UNLOCK", "SEND",
            "BECOME", "PUT", "CHANGE", "PLAY", "CYCLE",
            "CHAR", "DIE", "BIND", "ALL", "OTHERS",
            "SELF", "NOT", "ALLIGNED", "CONTACT", "BLOCKED",
            "ENERGIZED", "ANY", "N", "NORTH", "S",
            "SOUTH", "E", "EAST", "W", "WEST",
            "I", "IDLE", "SEEK", "FLOW", "RND",
            "RNDNS", "RNDNE", "CW", "CCW", "RNDP",
            "OPP", "BLUE", "GREEN", "CYAN", "RED",
            "PURPLE", "YELLOW", "WHITE", "HEALTH", "AMMO",
            "GEMS", "TORCHES", "SCORE", "TIME", "Z",
            "TOUCH", "THUD", "BOMBED", "ENERGIZE", "ENTER",
            "HINT", "SHOT"
        }

        Protected Shared ReadOnly OOPCodeSetBQ() As String = {
            "THEN", "GO", "TRY", "WALK", "SET",
            "CLEAR", "IF", "SHOOT", "THROWSTAR", "GIVE",
            "TAKE", "STP", "OHHDREA", "IDLE", "RESTART",
            "ZAP", "RESTORE", "LOCK", "UNLOCK", "SEND",
            "BECOME", "PUT", "CHANGE", "MUZK", "SPEED",
            "CHAR", "DIE", "BIND", "ALL", "OTHERS",
            "SELF", "NOT", "ALLIGNED", "CONTACT", "BLOCKED",
            "ENERGIZED", "ANY", "N", "NORTH", "S",
            "SOUTH", "E", "EAST", "W", "WEST",
            "I", "IDLE", "SEEK", "FLOW", "RND",
            "RNDNS", "RNDNE", "CW", "CCW", "RNDP",
            "OPP", "BLUE", "GREEN", "CYAN", "RED",
            "PURPLE", "YELLOW", "WHITE", "HEALTH", "AMMO",
            "GEMS", "TORCHES", "SCORE", "TIME", "Z",
            "TCTVT", "THUD", "BOMBED", "ENERGIZE", "ENTER",
            "HINT", "SHOT"
        }

        Protected OOPCodeSet() As String = OOPCodeSetNormal

        Protected Structure Kind
            Public Element As Element
            Public Color As Integer
        End Structure

        Protected Class OOPState
            Public DeathTile As Tile
            Public Previous As Integer
            Public Moved As Boolean
            Public Died As Boolean
            Public Commands As Integer
            Public Finished As Boolean
            Public AutoNextLine As Boolean
            Public Message As List(Of String)
            Public Repeat As Boolean
            Public Thing As Thing
        End Class

        Protected Structure CodeSearchResult
            Public Target As Thing
            Public Instruction As Integer
            Public Sub New(ByVal NewIndex As Thing, ByVal NewInst As Integer)
                Target = NewIndex
                Instruction = NewInst
            End Sub
        End Structure

        Public Sub SetCodeSet(ByVal CodeSet As OOPCodeSets)
            Select Case CodeSet
                Case OOPCodeSets.Normal
                    OOPCodeSet = OOPCodeSetNormal
                Case OOPCodeSets.BQuest
                    OOPCodeSet = OOPCodeSetBQ
            End Select
        End Sub

        Protected Function SendLabel(ByVal Sender As Thing, ByVal Label As String, ByVal Force As Boolean, ByVal NotSelfIfLocked As Boolean) As Boolean
            'external is if sender is >= 0

            Dim ModifiedSelf As Boolean = False
            Dim Index As Integer = 0
            Dim SearchOffset As Integer = 0
            Dim Locked As Boolean
            Dim Results() As CodeSearchResult

            Results = SendSearch(Sender, Label, Chr(13) & ":")

            For Index = 0 To Results.Length - 1
                With Results(Index).Target
                    If (Parameters.MultiMovement) Then
                        Locked = (.P3 <> 0)
                    Else
                        Locked = (.P2 <> 0)
                    End If
                    If (Not Locked) OrElse (Force = True) OrElse (Results(Index).Target.Equals(Sender) AndAlso (NotSelfIfLocked = False)) Then
                        If (Results(Index).Target.Equals(Sender)) Then
                            ModifiedSelf = True
                        End If
                        .CurrentInstruction = Results(Index).Instruction
                    End If
                End With
            Next

            Return ModifiedSelf
        End Function

        Protected Function SendSearch(ByVal Thing As Thing, ByVal Label As String, ByVal Prefix As String) As CodeSearchResult()
            Dim Index As Integer
            Dim Ret As New List(Of CodeSearchResult)
            Dim Min As Integer = 0
            Dim Max As Integer = ThingCount - 1
            Dim Target As String
            Dim TargetExists As Boolean
            Dim Code As CodeContainer
            Dim Offset As Integer

            TargetExists = Label.Contains(":")
            If (TargetExists) Then
                Target = Label.Substring(0, Label.IndexOf(":"))
                Label = Label.Substring(Label.IndexOf(":") + 1)
            Else
                Target = ""
                Min = CurrentBoard.Things.IndexOf(Thing)
                Max = Min
            End If

            For Index = Min To Max
                If (EvalSend(Thing, CurrentBoard.Things(Index), Target)) Then
                    Code = CurrentBoard.Things(Index).Code
                    If (Not (Code Is Nothing)) Then
                        Offset = OOPSearch(CurrentBoard.Things(Index), Prefix & Label)
                        If (Offset >= 0) Then
                            Ret.Add(New CodeSearchResult(CurrentBoard.Things(Index), Offset))
                        End If
                    End If
                End If
            Next

            Return Ret.ToArray
        End Function

        Private Function OOPSearch(ByVal ThingData As Thing, ByVal SearchString As String) As Integer
            Dim Offset As Integer
            Dim CodeOffset As Integer
            Dim Index As Integer
            Dim SearchOffset As Integer
            Dim SearchLength As Integer
            Dim SearchCode As String
            Dim CodeLength As Integer

            SearchLength = SearchString.Length
            SearchCode = ThingData.Code.Code
            CodeLength = SearchCode.Length

            Do
                If (CodeOffset > CodeLength) Then
                    Offset = -1
                    Exit Do
                End If
                SearchOffset = CodeOffset
                For Index = 1 To SearchLength
                    If (ReadOOPByte(ThingData, SearchOffset).ToUpper = SearchString.Substring(Index - 1, 1).ToUpper) Then
                        Continue For
                    Else
                        CodeOffset += 1
                        Continue Do
                    End If
                Next
                Select Case ReadOOPByte(ThingData, SearchOffset).ToUpper
                    Case "A" To "Z", "_"
                    Case Else
                        Offset = CodeOffset
                        Exit Do
                End Select
                CodeOffset += 1
            Loop
            Return Offset
        End Function

        Protected Function EvalSend(ByVal Sender As Thing, ByVal Receiver As Thing, ByVal Target As String) As Boolean
            Dim Code As CodeContainer

            If (Target = OOPCodeSet(28)) Then
                'ALL
                Return (True)
            ElseIf (Target = OOPCodeSet(29)) Then
                'OTHERS
                Return (Not Sender.Equals(Receiver))
            ElseIf (Target = "") OrElse (Target = OOPCodeSet(30)) Then
                'SELF/none
                Return (Sender.Equals(Receiver))
            Else
                Code = Receiver.Code
                If (Code Is Nothing) Then
                    Return (False)
                Else
                    Return ((ReadOOPByte(Receiver, 0) = "@") AndAlso (ReadOOPWord(Receiver, 1).ToUpper = Target))
                End If
            End If
        End Function

        Protected Function ReadOOPByte(ByVal Thing As Thing, ByRef Offset As Integer) As String
            Dim Ret As String
            Dim Code As CodeContainer

            Code = Thing.Code
            If (Code Is Nothing) OrElse (Offset >= Code.Code.Length) Then
                Ret = ""
            Else
                Ret = Code.Code.Substring(Offset, 1)
                Offset += 1
            End If
            Return Ret
        End Function

        Protected Function ReadOOPWord(ByVal Thing As Thing, ByRef Offset As Integer) As String
            Dim Ch As String
            Dim Ret As String = ""

            Do
                Ch = ReadOOPByte(Thing, Offset)
            Loop While (Ch = " ")
            Ch = Ch.ToUpper
            If (Ch < "0") OrElse (Ch > "9") Then
                Do
                    Select Case Ch
                        Case "A" To "Z", "0" To "9", ":", "_"
                            Ret &= Ch
                        Case Else
                            Exit Do
                    End Select
                    Ch = ReadOOPByte(Thing, Offset).ToUpper
                Loop
            End If
            If (Offset > 0) Then
                Offset -= 1
            End If
            Return Ret
        End Function

        Protected Function ReadOOPNumber(ByVal Thing As Thing, ByRef Offset As Integer) As Integer
            Dim Ch As String
            Dim Ret As String = ""
            Do
                Ch = ReadOOPByte(Thing, Offset)
            Loop While (Ch = " ")
            Do
                If (Ch >= "0") AndAlso (Ch <= "9") Then
                    Ret &= Ch
                    Ch = ReadOOPByte(Thing, Offset)
                Else
                    Exit Do
                End If
            Loop
            If (Offset > 0) Then
                Offset -= 1
            End If
            If (Ret = "") Then
                Return -1
            Else
                Return CInt(Ret)
            End If
        End Function

        Protected Function ReadOOPDirection(ByVal Thing As Thing, ByRef Offset As Integer) As Vector
            Dim V As Vector
            If (Not EvalDirection(Thing, Offset, V)) Then
                OOPError("Bad direction")
            End If
            Return V
        End Function

        Protected Function ReadOOPLine(ByVal Thing As Thing, ByRef Offset As Integer) As String
            Dim Ret As String = ""
            Dim Ch As String
            Do
                Ch = ReadOOPByte(Thing, Offset)
                If (Ch = "") OrElse (Ch = Chr(0)) OrElse (Ch = Chr(13)) Then
                    Exit Do
                End If
                Ret &= Ch
            Loop
            Return Ret
        End Function

        Protected Function EvalCondition(ByVal Thing As Thing, ByRef CurrentInst As Integer) As Boolean
            Dim V As Vector = Nothing
            Dim K As Kind = Nothing
            Dim W As String
            W = ReadOOPWord(Thing, CurrentInst)

            Select Case W
                Case OOPCodeSet(31)
                    'NOT
                    Return (Not EvalCondition(Thing, CurrentInst))
                Case OOPCodeSet(32)
                    'ALLIGNED
                    With Thing
                        Return ((.X = PlayerX) OrElse (.Y = PlayerY))
                    End With
                Case OOPCodeSet(33)
                    'CONTACT
                    With Thing
                        Return (((.Y - PlayerY) ^ 2) + ((.X - PlayerX) ^ 2) = 1)
                    End With
                Case OOPCodeSet(34)
                    'BLOCKED
                    With Thing
                        V = ReadOOPDirection(Thing, CurrentInst)
                        Return (Not ElementAt(.X + V.X, .Y + V.Y).Floor)
                    End With
                Case OOPCodeSet(35)
                    'ENERGIZED
                    Return (World.EnergyCycles > 0)
                Case OOPCodeSet(36)
                    'ANY
                    If (EvalKind(Thing, CurrentInst, K)) Then
                        Return IsAny(K)
                    Else
                        OOPError("Bad object kind")
                        Return False
                    End If
                Case Else
                    Return FindFlag(W)
            End Select
        End Function

        Protected Function EvalDirection(ByVal Thing As Thing, ByRef CurrentInst As Integer, ByRef ReturnVector As Vector) As Boolean
            Dim Success As Boolean
            Select Case ReadOOPWord(Thing, CurrentInst)
                Case OOPCodeSet(37), OOPCodeSet(38)
                    'N, NORTH
                    ReturnVector = Vector.North
                    Success = True
                Case OOPCodeSet(39), OOPCodeSet(40)
                    'S, SOUTH
                    ReturnVector = Vector.South
                    Success = True
                Case OOPCodeSet(41), OOPCodeSet(42)
                    'E, EAST
                    ReturnVector = Vector.East
                    Success = True
                Case OOPCodeSet(43), OOPCodeSet(44)
                    'W, WEST
                    ReturnVector = Vector.West
                    Success = True
                Case OOPCodeSet(45), OOPCodeSet(46)
                    'I, IDLE
                    ReturnVector = Vector.Idle
                    Success = True
                Case OOPCodeSet(47)
                    'SEEK
                    With Thing
                        ReturnVector = Seek(.X, .Y)
                    End With
                    Success = True
                Case OOPCodeSet(48)
                    'FLOW
                    With Thing
                        ReturnVector = .Flow
                    End With
                    Success = True
                Case OOPCodeSet(49)
                    'RND
                    ReturnVector = Rnd()
                    Success = True
                Case OOPCodeSet(50)
                    'RNDNS
                    ReturnVector = RndNS()
                    Success = True
                Case OOPCodeSet(51)
                    'RNDNE
                    ReturnVector = RndNE()
                    Success = True
                Case OOPCodeSet(52)
                    'CW
                    Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                    ReturnVector = New Vector(-ReturnVector.Y, ReturnVector.X)
                Case OOPCodeSet(53)
                    'CCW
                    Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                    ReturnVector = New Vector(ReturnVector.Y, -ReturnVector.X)
                Case OOPCodeSet(54)
                    'RNDP
                    Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                    ReturnVector = RndP(ReturnVector)
                Case OOPCodeSet(55)
                    'OPP
                    Success = EvalDirection(Thing, CurrentInst, ReturnVector)
                    ReturnVector = ReturnVector.Opp
            End Select
            Return Success
        End Function

        Protected Function EvalKind(ByVal Thing As Thing, ByRef CurrentInst As Integer, ByRef ReturnKind As Kind) As Boolean
            Dim ElementIndex As Integer
            Dim Word As String

            ReturnKind = New Kind
            ReturnKind.Color = 0
            ReturnKind.Element = Nothing
            Do
                Word = ReadOOPWord(Thing, CurrentInst)
                Select Case Word
                    Case OOPCodeSet(56)
                        'BLUE
                        ReturnKind.Color = 9
                        Continue Do
                    Case OOPCodeSet(57)
                        'GREEN
                        ReturnKind.Color = 10
                        Continue Do
                    Case OOPCodeSet(58)
                        'CYAN
                        ReturnKind.Color = 11
                        Continue Do
                    Case OOPCodeSet(59)
                        'RED
                        ReturnKind.Color = 12
                        Continue Do
                    Case OOPCodeSet(60)
                        'PURPLE
                        ReturnKind.Color = 13
                        Continue Do
                    Case OOPCodeSet(61)
                        'YELLOW
                        ReturnKind.Color = 14
                        Continue Do
                    Case OOPCodeSet(62)
                        'WHITE
                        ReturnKind.Color = 15
                        Continue Do
                    Case Else
                        For ElementIndex = 0 To Library.Count - 1
                            If (Word = Library.GetElementFromIndex(ElementIndex).OOPString) Then
                                ReturnKind.Element = Library.GetElementFromIndex(ElementIndex)
                                Exit For
                            End If
                        Next
                End Select
                Exit Do
            Loop
            Return (Not (ReturnKind.Element Is Nothing))
        End Function

        Protected Function IsAny(ByVal K As Kind) As Boolean
            Dim X As Integer
            Dim Y As Integer
            Dim H As Integer = Height - 1
            Dim W As Integer = Width - 1
            Dim T As Tile

            For Y = 0 To H
                For X = 0 To W
                    T = TileAt(X, Y)
                    If (T.Element.ID = K.Element.ID) Then
                        If (K.Color = 0) OrElse (K.Color = ColorMatch(T)) Then
                            Return True
                        End If
                    End If
                Next
            Next
            Return False
        End Function

        Protected Function ColorMatch(ByVal Tile As Tile) As Integer
            If (Tile.Element.Color < &HF0) Then
                Return (Tile.Element.Color And 7)
            ElseIf (Tile.Element.Color = &HFE) Then
                Return (((Tile.Color >> 4) And &HF) + 8)
            End If
            Return (Tile.Color And &HF)
        End Function

        Protected Function ExecuteOOP(ByVal Index As Integer, ByVal Instruction As Integer, ByVal Title As String)
            Dim OOP As New OOPState
            Dim Command As String
            Dim Prefix As String
            Dim V As Vector
            Dim TargetX As Integer
            Dim TargetY As Integer
            Dim Val As Integer
            Dim K As Kind
            Dim MoveCount As Integer
            Dim Message As String
            Dim MessageTitle As String
            Dim Thing As Thing = GetFixedThing(Index)

            OOP.DeathTile = New Tile(Library.GetElementFromID(ElementID.Empty), &HF)
            OOP.Message = New List(Of String)
            OOP.Thing = Thing
            OOP.Thing.CurrentInstruction = Instruction

            Do
                If (OOP.Thing.Code Is Nothing) Then
                    OOP.Thing.CurrentInstruction = -1
                    Return False
                End If
                OOP.Previous = OOP.Thing.CurrentInstruction
                OOP.AutoNextLine = True

                Prefix = ReadOOPByte(OOP.Thing, OOP.Thing.CurrentInstruction)
                Select Case Prefix
                    Case ":"
                        ReadOOPLine(OOP.Thing, OOP.Thing.CurrentInstruction)
                    Case "'"
                        ReadOOPLine(OOP.Thing, OOP.Thing.CurrentInstruction)
                    Case "@"
                        ReadOOPLine(OOP.Thing, OOP.Thing.CurrentInstruction)
                    Case "/", "?"
                        If (Prefix = "/") Then
                            OOP.Repeat = True
                        End If
                        If (EvalDirection(OOP.Thing, OOP.Thing.CurrentInstruction, V)) Then
                            If (Parameters.MultiMovement) Then
                                MoveCount = ReadOOPNumber(OOP.Thing, OOP.Thing.CurrentInstruction)
                                If (MoveCount < 0) Then
                                    MoveCount = 1
                                End If
                                If (Prefix = "?") Then
                                    MoveCount = (-MoveCount) And 255
                                End If
                                OOP.Thing.P2 = MoveCount
                                OOP.Thing.Vector = V
                                OOP.Repeat = False
                            Else
                                If (V.X = 0) AndAlso (V.Y = 0) Then
                                    OOP.Repeat = False
                                Else
                                    TargetX = OOP.Thing.X + V.X
                                    TargetY = OOP.Thing.Y + V.Y
                                    If (Not ElementAt(TargetX, TargetY).Floor) Then
                                        Push(TargetX, TargetY, V)
                                    End If
                                    If (ElementAt(TargetX, TargetY).Floor) Then
                                        MoveThing(Index, TargetX, TargetY)
                                        OOP.Repeat = False
                                    End If
                                End If
                            End If
                            If (ReadOOPByte(OOP.Thing, OOP.Thing.CurrentInstruction) <> Chr(13)) Then
                                OOP.Thing.CurrentInstruction -= 1
                            End If
                            OOP.Moved = True
                        End If
                    Case "#"
                        Do
                            Command = ReadOOPWord(OOP.Thing, OOP.Thing.CurrentInstruction)
                            If (Command = OOPCodeSet(0)) Then
                                'THEN
                                Command = ReadOOPWord(OOP.Thing, OOP.Thing.CurrentInstruction)
                            End If
                            If (Command = "") Then
                                Exit Do
                            End If
                            OOP.Commands += 1
                            Select Case Command
                                Case OOPCodeSet(1)
                                    'GO
                                    V = ReadOOPDirection(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    TargetX = Thing.X + V.X
                                    TargetY = Thing.Y + V.Y
                                    If (Not ElementAt(TargetX, TargetY).Floor) Then
                                        Push(TargetX, TargetY, V)
                                        If (Parameters.KoopoBug) Then
                                            OOP.Thing = GetFixedThing(Index)
                                        Else
                                            Index = CurrentBoard.Things.IndexOf(OOP.Thing)
                                        End If
                                    End If
                                    If (ElementAt(TargetX, TargetY).Floor) Then
                                        MoveThing(Index, TargetX, TargetY)
                                    Else
                                        OOP.Repeat = True
                                    End If
                                    OOP.Moved = True
                                Case OOPCodeSet(2)
                                    'TRY
                                    V = ReadOOPDirection(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    TargetX = Thing.X + V.X
                                    TargetY = Thing.Y + V.Y
                                    If (Not ElementAt(TargetX, TargetY).Floor) Then
                                        Push(TargetX, TargetY, V)
                                        If (Parameters.KoopoBug) Then
                                            OOP.Thing = GetFixedThing(Index)
                                        Else
                                            Index = CurrentBoard.Things.IndexOf(OOP.Thing)
                                        End If
                                    End If
                                    If (ElementAt(TargetX, TargetY).Floor) Then
                                        MoveThing(Index, TargetX, TargetY)
                                        OOP.Moved = True
                                    Else
                                        OOP.Thing.CurrentInstruction += 1
                                        Continue Do
                                    End If
                                Case OOPCodeSet(3)
                                    'WALK
                                    V = ReadOOPDirection(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    With OOP.Thing
                                        .XStep = V.X
                                        .YStep = V.Y
                                    End With
                                Case OOPCodeSet(4)
                                    'SET
                                    SetFlag(ReadOOPWord(OOP.Thing, OOP.Thing.CurrentInstruction))
                                Case OOPCodeSet(5)
                                    'CLEAR
                                    ClearFlag(ReadOOPWord(OOP.Thing, OOP.Thing.CurrentInstruction))
                                Case OOPCodeSet(6)
                                    'IF
                                    If (EvalCondition(OOP.Thing, OOP.Thing.CurrentInstruction)) Then
                                        Continue Do
                                    End If
                                Case OOPCodeSet(7)
                                    'SHOOT
                                    V = ReadOOPDirection(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    If SpawnProjectile(Library.GetElementFromID(ElementID.Bullet), OOP.Thing.X, OOP.Thing.Y, V, True) Then
                                        'playsound(2, enemy bullet)
                                    End If
                                    OOP.Moved = True
                                Case OOPCodeSet(8)
                                    'THROWSTAR
                                    V = ReadOOPDirection(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    TargetX = OOP.Thing.X + V.X
                                    TargetY = OOP.Thing.Y + V.Y
                                    SpawnProjectile(Library.GetElementFromID(ElementID.Star), TargetX, TargetY, V, True)
                                    OOP.Moved = True
                                Case OOPCodeSet(9), OOPCodeSet(10)
                                    'GIVE, TAKE
                                    If (Not ExecuteOOPGive(OOP.Thing, OOP, Command)) Then
                                        OOP.Thing.CurrentInstruction += 1
                                        Continue Do
                                    End If
                                Case OOPCodeSet(11)
                                    'END
                                    OOP.Thing.CurrentInstruction = -1
                                    OOP.AutoNextLine = False
                                    OOP.Finished = True
                                Case OOPCodeSet(12)
                                    'ENDGAME
                                    World.Health = 0
                                Case OOPCodeSet(13)
                                    'IDLE
                                    OOP.Moved = True
                                Case OOPCodeSet(14)
                                    'RESTART
                                    OOP.Thing.CurrentInstruction = 0
                                    OOP.AutoNextLine = False
                                Case OOPCodeSet(15)
                                    'ZAP
                                    ExecuteOOPZap(OOP.Thing, OOP)
                                Case OOPCodeSet(16)
                                    'RESTORE
                                    ExecuteOOPRestore(OOP.Thing, OOP)
                                Case OOPCodeSet(17)
                                    'LOCK
                                    If (Parameters.MultiMovement) Then
                                        OOP.Thing.P3 = 1
                                    Else
                                        OOP.Thing.P2 = 1
                                    End If
                                Case OOPCodeSet(18)
                                    'UNLOCK
                                    If (Parameters.MultiMovement) Then
                                        OOP.Thing.P3 = 0
                                    Else
                                        OOP.Thing.P2 = 0
                                    End If
                                Case OOPCodeSet(19)
                                    'SEND
                                    If (SendLabel(OOP.Thing, ReadOOPWord(OOP.Thing, OOP.Thing.CurrentInstruction), False, False)) Then
                                        OOP.AutoNextLine = False
                                    End If
                                Case OOPCodeSet(20)
                                    'BECOME
                                    K = Nothing
                                    If EvalKind(OOP.Thing, OOP.Thing.CurrentInstruction, K) Then
                                        OOP.DeathTile = New Tile(K.Element, K.Color)
                                        OOP.Died = True
                                    Else
                                        OOPError("Bad #BECOME")
                                    End If
                                Case OOPCodeSet(21)
                                    'PUT
                                    V = ReadOOPDirection(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    K = Nothing
                                    If ((V.X = 0) AndAlso (V.Y = 0)) OrElse (Not EvalKind(OOP.Thing, OOP.Thing.CurrentInstruction, K)) Then
                                        OOPError("Bad #PUT")
                                    Else
                                        TargetX = OOP.Thing.X + V.X
                                        TargetY = OOP.Thing.Y + V.Y
                                        If (TargetX >= 0) AndAlso (TargetX < Width) AndAlso (TargetY >= 0) AndAlso (TargetY < Height) Then
                                            If (Not ElementAt(TargetX, TargetY).Floor) Then
                                                Push(TargetX, TargetY, V)
                                            End If
                                            PlotTile(TargetX, TargetY, New Tile(K.Element, K.Color))
                                        End If
                                    End If
                                    If (Parameters.KoopoBug) Then
                                        OOP.Thing = GetFixedThing(Index)
                                    Else
                                        Index = CurrentBoard.Things.IndexOf(OOP.Thing)
                                    End If
                                Case OOPCodeSet(22)
                                    'CHANGE
                                    ExecuteOOPChange(OOP.Thing, OOP)
                                    If (Parameters.KoopoBug) Then
                                        OOP.Thing = GetFixedThing(Index)
                                    Else
                                        Index = CurrentBoard.Things.IndexOf(OOP.Thing)
                                    End If
                                Case OOPCodeSet(23)
                                    'PLAY
                                Case OOPCodeSet(24)
                                    'CYCLE
                                    Val = ReadOOPNumber(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    If (Val > 0) Then
                                        OOP.Thing.Cycle = Val
                                    End If
                                Case OOPCodeSet(25)
                                    'CHAR
                                    Val = ReadOOPNumber(OOP.Thing, OOP.Thing.CurrentInstruction)
                                    If (Val >= 0) AndAlso (Val <= 255) Then
                                        OOP.Thing.P1 = Val
                                    End If
                                    UpdateAt(OOP.Thing.X, OOP.Thing.Y)
                                Case OOPCodeSet(26)
                                    'DIE
                                    OOP.Died = True
                                    OOP.DeathTile = New Tile(Library.GetElementFromID(ElementID.Empty), &HF)
                                Case OOPCodeSet(27)
                                    'BIND
                                    ExecuteOOPBind(OOP.Thing, OOP)
                                Case Else
                                    'label
                                    If SendLabel(OOP.Thing, Command, False, False) Then
                                        OOP.AutoNextLine = False
                                    ElseIf (Not Command.Contains(":")) Then
                                        'OOPError("Bad command " & Command)
                                    End If
                            End Select
                            If (Not OOP.Died) AndAlso (OOP.AutoNextLine) AndAlso (OOP.Thing.CurrentInstruction > -1) Then
                                ReadOOPLine(OOP.Thing, OOP.Thing.CurrentInstruction)
                            End If
                            Exit Do
                        Loop
                        If (Command = "") Then
                            Continue Do
                        End If
                    Case Chr(13)
                        If (OOP.Message.Count > 0) Then
                            OOP.Message.Add(Prefix)
                        End If
                    Case "", Chr(0)
                        OOP.Finished = True
                    Case Else
                        OOP.Message.Add(Prefix & ReadOOPLine(OOP.Thing, OOP.Thing.CurrentInstruction))
                End Select
                If (Parameters.KoopoBug) Then
                    OOP.Thing = GetFixedThing(Index)
                Else
                    Index = CurrentBoard.Things.IndexOf(OOP.Thing)
                End If
            Loop Until (OOP.Finished) OrElse (OOP.Moved) OrElse (OOP.Repeat) OrElse (OOP.Died) OrElse (OOP.Commands > &H20) OrElse (OOP.Thing.CurrentInstruction < 0)
            If (OOP.Repeat) Then
                OOP.Thing.CurrentInstruction = OOP.Previous
            End If
            If (OOP.Thing.Code Is Nothing) OrElse (OOP.Thing.CurrentInstruction >= OOP.Thing.Code.Code.Length) Then
                OOP.Thing.CurrentInstruction = -1
            End If
            If (OOP.Message.Count > Parameters.MessageLines) Then
                Message = ""
                If (ReadOOPByte(Thing, 0) = "@") Then
                    MessageTitle = ReadOOPLine(Thing, 1)
                Else
                    MessageTitle = Title
                End If
                For Val = 0 To OOP.Message.Count - 1
                    Message &= Chr(13) & OOP.Message(Val)
                Next
                If (Message <> "") Then
                    Message = Message.Remove(0, 1)
                    ExecuteScroll(Thing, MessageTitle, Message)
                End If
            ElseIf (OOP.Message.Count > 0) Then
                Message = ""
                For Val = 0 To OOP.Message.Count - 1
                    Message &= Chr(13) & OOP.Message(Val)
                Next
                If (Message <> "") Then
                    Message = Message.Remove(0, 1)
                    SetMessage(&HC8, Message)
                End If
            End If
            If (OOP.Died) Then
                TargetX = OOP.Thing.X
                TargetY = OOP.Thing.Y
                If (Parameters.PreKillObject) Then
                    HarmThing(ThingIndexAt(TargetX, TargetY))
                End If
                PlotTile(TargetX, TargetY, OOP.DeathTile)
            End If
            Thing.CurrentInstruction = OOP.Thing.CurrentInstruction
            Return (OOP.Died)
        End Function

        Protected Sub ExecuteOOPBind(ByVal Thing As Thing, ByRef OOP As OOPState)
            Dim Target As String
            Dim Max As Integer = ThingCount - 1
            Dim Index As Integer

            Target = ReadOOPWord(Thing, OOP.Thing.CurrentInstruction)
            For Index = 0 To ThingCount - 1
                If EvalSend(Thing, CurrentBoard.Things(Index), Target) Then
                    With CurrentBoard.Things(Index)
                        Thing.Code = .Code
                        Thing.CurrentInstruction = 0
                        Thing.Length = .Length
                        OOP.Thing.CurrentInstruction = 0
                    End With
                    Exit For
                End If
            Next
        End Sub

        Protected Sub ExecuteOOPChange(ByVal Thing As Thing, ByRef OOP As OOPState)
            Dim SourceKind As Kind = Nothing
            Dim TargetKind As Kind = Nothing
            Dim X As Integer
            Dim Y As Integer
            Dim W As Integer = Width - 1
            Dim H As Integer = Height - 1
            Dim T As Tile

            If (Not EvalKind(Thing, OOP.Thing.CurrentInstruction, SourceKind)) OrElse (Not EvalKind(Thing, OOP.Thing.CurrentInstruction, TargetKind)) Then
                OOPError("Bad #CHANGE")
            Else
                If (TargetKind.Color = 0) AndAlso (TargetKind.Element.Color < &HF0) Then
                    TargetKind.Color = TargetKind.Element.Color
                End If
                For Y = 0 To H
                    For X = 0 To W
                        T = TileAt(X, Y)
                        If (T.Element.ID = SourceKind.Element.ID) Then
                            If (SourceKind.Color = 0) OrElse (SourceKind.Color = ColorMatch(T)) Then
                                PlotTile(X, Y, New Tile(TargetKind.Element, TargetKind.Color))
                            End If
                        End If
                    Next
                Next
            End If
        End Sub

        Protected Function ExecuteOOPGive(ByVal Thing As Thing, ByRef OOP As OOPState, ByVal Command As String) As Boolean
            Dim Result As Boolean = False
            Dim Item As String
            Dim Count As Integer

            Item = ReadOOPWord(Thing, OOP.Thing.CurrentInstruction)
            Count = ReadOOPNumber(Thing, OOP.Thing.CurrentInstruction)
            If (Count > 0) Then
                If (Command = OOPCodeSet(10)) Then
                    Count = -Count
                End If
                Select Case Item
                    Case "HEALTH"
                        Result = ExecuteOOPGiveAttempt(World.Health, Count)
                    Case "AMMO"
                        Result = ExecuteOOPGiveAttempt(World.Ammo, Count)
                    Case "GEMS"
                        Result = ExecuteOOPGiveAttempt(World.Gems, Count)
                    Case "TORCHES"
                        If (Parameters.TorchesEnabled) Then
                            Result = ExecuteOOPGiveAttempt(World.Torches, Count)
                        End If
                    Case "SCORE"
                        Result = ExecuteOOPGiveAttempt(World.Score, Count)
                    Case "TIME"
                        Result = ExecuteOOPGiveAttempt(World.TimePassed, Count)
                    Case "Z"
                        If (Parameters.StonesEnabled) Then
                            Result = ExecuteOOPGiveAttempt(World.Stones, Count)
                        End If
                End Select
            End If
            StatusUpdate()
            Return Result
        End Function

        Protected Function ExecuteOOPGiveAttempt(ByRef Target As Integer, ByVal Amount As Integer) As Boolean
            Dim Val As Integer
            Val = Target + Amount
            If (Val >= 0) Then
                Target += Amount
                Return True
            End If
            Return False
        End Function

        Protected Sub ExecuteOOPRestore(ByVal Thing As Thing, ByRef OOP As OOPState)
            Dim Target As String = ReadOOPWord(Thing, OOP.Thing.CurrentInstruction)
            Dim SearchOffset As Integer = 0
            Dim SendResult() As CodeSearchResult
            Dim Code As CodeContainer
            Dim Offset As Integer

            Do
                SendResult = SendSearch(Thing, Target, Chr(13) & "'")
                If (SendResult.Length > 0) Then
                    With SendResult(0)
                        Offset = .Instruction
                        Code = .Target.Code
                        Do
                            Code.Code = Code.Code.Remove(Offset + 1, 1).Insert(Offset + 1, ":")
                            Offset = OOPSearch(.Target, Chr(13) & "'" & Target)
                        Loop While (Offset > 0)
                    End With
                Else
                    Exit Do
                End If
            Loop
        End Sub

        Protected Sub ExecuteOOPZap(ByVal Thing As Thing, ByRef OOP As OOPState)
            Dim Target As String = ReadOOPWord(Thing, OOP.Thing.CurrentInstruction)
            Dim SearchOffset As Integer = 0
            Dim Results() As CodeSearchResult = SendSearch(Thing, Target, Chr(13) & ":")
            Dim Index As Integer
            Dim Code As CodeContainer
            Dim Offset As Integer

            For Index = 0 To Results.Length - 1
                With Results(Index)
                    Code = .Target.Code
                    Offset = .Instruction + 1
                    Code.Code = Code.Code.Remove(Offset, 1).Insert(Offset, "'")
                End With
            Next
        End Sub

        Protected Sub ExecuteScroll(ByVal Owner As Thing, ByVal Title As String, ByVal Text As String)
            If Not (Scroll.Visible) Then
                ScrollOwner = Owner
                Scroll.Text = Text
                Scroll.Title = Title
                Scroll.Open()
            End If
        End Sub

        Protected Sub ExecuteScrollClose(ByVal Owner As Thing, ByVal Index As Integer)
            Dim Message() As String = Scroll.Text.Split(Chr(13))
            Dim Label As String

            If (Message(Index) <> "") AndAlso (Message(Index).Substring(0, 1) = "!") AndAlso (Message(Index).IndexOf(";") > 0) Then
                Label = Message(Index).Substring(1)
                Label = Label.Substring(0, Label.IndexOf(";"))
                SendLabel(Owner, Label.ToUpper, True, False)
            End If
        End Sub

        Protected Function FindFlag(ByVal Name As String) As Boolean
            Dim Result As Boolean
            If (World.Flags.Contains(Name)) Then
                Result = True
            ElseIf (Name = "") AndAlso (World.Flags.Count < Parameters.MaxFlags) Then
                Result = True
            End If
            Return Result
        End Function

        Protected Sub SetFlag(ByVal Name As String)
            If (Not FindFlag(Name)) Then
                If (World.Flags.Count < Parameters.MaxFlags) Then
                    World.Flags.Add(Name.ToUpper)
                Else
                    World.Flags(Parameters.MaxFlags - 1) = Name
                End If
            End If
        End Sub

        Protected Sub ClearFlag(ByVal Name As String)
            Dim Index As Integer
            Index = World.Flags.IndexOf(Name)
            If (Index >= 0) Then
                World.Flags.RemoveAt(Index)
            End If
        End Sub

        Protected Sub OOPError(ByVal Message As String)
            SetMessage(&HC8, "ERR: " & Message)
            'playsound(5, oop error)
        End Sub

        Protected Function GetFixedThing(ByVal Index As Integer) As Thing
            Return CurrentBoard.Things(Index)
        End Function

    End Class
End Namespace